"""
NRV-:class:`.fascicle_results` handling.
"""

from ...backend._NRV_Results import sim_results
from ...backend._log_interface import pass_info, rise_warning, rise_error
from ...fmod._electrodes import is_FEM_electrode
from ...utils._units import nm, convert, from_nrv_unit, to_nrv_unit
from ...utils._misc import membrane_capacitance_from_model, compute_complex_admitance
import matplotlib.pyplot as plt
import numpy as np


def number_in_str(s: str) -> bool:
    return any(i.isdigit() for i in s)


recognized_axon_types = ["all", "myelinated", "unmyelinated"]


class fascicle_results(sim_results):
    """ """

    def __init__(self, context=None):
        super().__init__(context)

    @property
    def n_ax(self):
        """
        Number of axons in the fascicle
        """
        return len(self.axons_diameter)

    def get_n_ax(self, ax_type: str = "all") -> int:
        """
        Number of myelinated axons in the fascicle
        """
        return len(self.get_axons_key(ax_type))

    def get_axons_key(self, ax_type: str = "all") -> list:
        """ """
        if ax_type not in recognized_axon_types:
            rise_error(
                f"Axon type specified not recognized. Recognized types are: {recognized_axon_types}"
            )
        all_keys = self.keys()
        axon_keys = [i for i in all_keys if ("axon" in i and number_in_str(i))]
        if ax_type != "all":
            if ax_type == "unmyelinated":
                axon_keys = [i for i in axon_keys if self[i].myelinated == False]
            else:
                axon_keys = [i for i in axon_keys if self[i].myelinated == True]
        return axon_keys

    def get_recruited_axons(
        self, ax_type: str = "all", normalize: bool = False,
        vm_key: str = "V_mem", t_start: float = None
    ) -> int | float:
        """
        Return the number or the ratio of recruited axons in the fascicle

        Parameters
        ----------
        ax_type : str, optional
            type of axon counted,possible options:
             - "all" (default)
             - "unmyelinated"
             - "myelinated"
        normalize : bool, optional
            if False the total number of recruited axons is returned, else the ratio is returned, by default False

        Returns
        -------
        int or float
            number of recruited axons
        """
        axons_keys = self.get_axons_key(ax_type)
        n_recr = 0
        for axon in axons_keys:
            if self[axon].is_recruited(vm_key, t_start):
                n_recr += 1
        if normalize:
            n_recr /= self.get_n_ax(ax_type)
        return n_recr

    def get_recruited_axons_greater_than(
        self, diam: float, ax_type: str = "all", normalize: bool = False,
        vm_key: str = "V_mem", t_start: float = None
    ) -> float:
        """
        Return the number or the ratio of recruited axons with a diameter greater than `diam` in the fascicle

        Parameters
        ----------
        ax_type : str, optional
            type of axon counted, possible options:
                - "all" (default)
                - "unmyelinated"
                - "myelinated"
        normalize : bool, optional
            if False the total number of recruited axons is returned, else the ratio is returned, by default False

        Returns
        -------
        int or float
            number of recruited axons
        """
        axons_keys = self.get_axons_key(ax_type)
        n_recr = 0
        n_tot = 0
        for axon in axons_keys:
            if self[axon].diameter > diam:
                n_tot += 1
                if self[axon].is_recruited(vm_key,t_start):
                    n_recr += 1
        if normalize:
            n_recr /= n_tot
        return n_recr

    def get_recruited_axons_lesser_than(
        self, diam: float, ax_type: str = "all", normalize: bool = False,
        vm_key: str = "V_mem", t_start: float = None
    ) -> float:
        """
        Return the number or the ratio of recruited axons with a diameter smaller than `diam` in the fascicle

        Parameters
        ----------
        ax_type : str, optional
            type of axon counted, possible options:
                - "all" (default)
                - "unmyelinated"
                - "myelinated"
        normalize : bool, optional
            if False the total number of recruited axons is returned, else the ratio is returned, by default False

        Returns
        -------
        int or float
            number of recruited axons
        """
        axons_keys = self.get_axons_key(ax_type)
        n_recr = 0
        n_tot = 0
        for axon in axons_keys:
            if self[axon].diameter < diam:
                n_tot += 1
                if self[axon].is_recruited(vm_key,t_start):
                    n_recr += 1
        if normalize:
            n_recr /= n_tot
        return n_recr

    def get_axons(self,vm_key: str = "V_mem", t_start: float = None) -> list:
        axons_keys = self.get_axons_key()
        axon_diam = []
        axon_type = []
        axon_y = []
        axon_z = []
        axon_recruited = []
        for axon in axons_keys:
            axon_recruited.append(self[axon].is_recruited(vm_key,t_start))
            axon_y.append(self[axon].y)
            axon_z.append(self[axon].z)
            axon_diam.append(self[axon].diameter)
            axon_type.append(self[axon].myelinated)
        return (axon_diam, axon_type, axon_y, axon_z, axon_recruited)

    def get_block_summary_axons(
        self, AP_start: float, freq: float = None, t_refractory: float = 1
    ) -> list:
        axons_keys = self.get_axons_key()
        axon_diam = []
        axon_type = []
        axon_y = []
        axon_z = []
        is_blocked = []
        n_onset = []
        for axon in axons_keys:
            self[axon].block_summary(AP_start, freq, t_refractory)

            is_blocked.append(self[axon].is_blocked)
            n_onset.append(self[axon].n_onset)
            axon_y.append(self[axon].y)
            axon_z.append(self[axon].z)
            axon_diam.append(self[axon].diameter)
            axon_type.append(self[axon].myelinated)
        return (axon_diam, axon_type, axon_y, axon_z, is_blocked, n_onset)

    # impeddance related methods
    def get_membrane_conductivity(
        self, x: float = 0, t: float = 0, unit: str = "S/cm**2", mem_th: float = 7 * nm
    ) -> np.array:
        """
        get the membran conductivity of each axon at a position x and a time t

        Parameters
        ----------
        x : float, optional
            x-position in um where to get the conductivity, by default 0
        t : float, optional
            simulation time in ms when to get the conductivity, by default 0
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm

        Note
        ----
        depending of the unit parameter this function either return :

            - the surface conductivity in [S]/([m]*[m]): from neuron simulation
            - the conductivity in [S]/[m]:  by multiplying surface conductivity by membrane thickness
        """

        g = []
        a_keys = self.get_axons_key()
        for key in a_keys:
            g_ = self[key].get_membrane_conductivity(x=x, t=t, unit=unit, mem_th=mem_th)
            if g_ is not None:
                g = np.concatenate((g, [g_]))
            else:
                return None
        return g

    def get_membrane_capacitance(
        self, unit: str = "uF/cm**2", mem_th: float = 7 * nm
    ) -> tuple[float]:
        """
        get the membrane capacitance or permitivity of unmyelinated and myelinated axons filling the ner

        Parameters
        ----------
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm

        Note
        ----
        depending of the unit parameter this function either return :

            - the surface capacitance in [S]/([m]*[m]): from neuron simulation
            - the permitivity in [S]/[m]:  by multiplying surface conductivity by membrane thickness
        """
        u_c_mem = membrane_capacitance_from_model(self.unmyelinated_param["model"])
        m_c_mem = membrane_capacitance_from_model(self.myelinated_param["model"])

        # Surface capacity in [F]/([m]*[m])
        if "2" in unit:
            return convert(u_c_mem, "S/cm**2", unit), convert(m_c_mem, "S/cm**2", unit)
        # permitivity in [F]/[m]
        else:
            u_c_mem *= from_nrv_unit(mem_th, "cm")
            m_c_mem *= from_nrv_unit(mem_th, "cm")
            return convert(u_c_mem, "S/cm", unit), convert(m_c_mem, "S/cm", unit)

    def get_membrane_complexe_admitance(
        self,
        f: float = 1.0,
        x: float = 0,
        t: float = 0,
        unit: str = "S/m",
        mem_th: float = 7 * nm,
    ) -> np.array:
        """
        get the membran complexe admitance of each axon at a position x and a time t for a given frequency

        Parameters
        ----------
        f : float or np.array, optional
            effective frequency in kHz, by default 1
        x : float, optional
            x-position in um where to get the conductivity, by default 0
        t : float, optional
            simulation time in ms when to get the conductivity, by default 0
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm
        """
        u_c, m_c = self.get_membrane_capacitance(mem_th=mem_th)
        eps = (self.axons_type * (m_c - u_c)) + u_c
        g = self.get_membrane_conductivity(x=x, t=t, mem_th=mem_th)
        f_mem = g / (2 * np.pi * eps)

        # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
        # [MHz] to convert to [kHz]
        f_mem = to_nrv_unit(f_mem, "MHz")

        Y = compute_complex_admitance(f=f, g=g, fc=f_mem)

        if "2" in unit:
            return convert(Y, "S/cm**2", unit)
        # permitivity in [F]/[m]
        else:
            Y *= from_nrv_unit(mem_th, "cm")
            return convert(Y, "S/cm", unit)

    def get_block_summary(
        self,
        AP_start: float,
        freq: float = None,
        t_refractory_m: float = 1,
        t_refractory_um: float = 1,
    ) -> None:
        """
        Get block characteristics (blocked, onset response, number of APs) for each axon of the fascicle

        Parameters
        ----------
        AP_start : float
            timestamp of the test pulse start, in ms.
        freq : float, optional
            Frequency of the stimulation, for KES block, by default None
        t_refractory_m : float, optional
            Axon refractory period for myelinated fibers, by default 1
        t_refractory_um : float, optional
            Axon refractory period for unmyelinated fibers, by default 1
        """
        axons_keys = self.get_axons_key()
        for axon in axons_keys:
            if self[axon].myelinated == True:
                self[axon].block_summary(AP_start, freq, t_refractory_m)
            else:
                self[axon].block_summary(AP_start, freq, t_refractory_um)

    ## Representation methods
    def plot_recruited_fibers(
        self,
        axes: plt.axes,
        contour_color: str = "k",
        myel_color: str = "r",
        unmyel_color: str = "b",
        num: bool = False,
    ) -> None:
        ## plot contour
        axes.add_patch(
            plt.Circle(
                (self.y_grav_center, self.z_grav_center),
                self.D / 2,
                color=contour_color,
                fill=False,
                linewidth=2,
            )
        )
        ## plot axons
        axon_diam, axon_type, axon_y, axon_z, axon_recruited = self.get_axons()
        for k, _ in enumerate(axon_diam):
            color = unmyel_color
            if axon_type[k]:
                color = myel_color
            alpha = 0.1
            if axon_recruited[k]:
                alpha = 1
            axes.add_patch(
                plt.Circle(
                    (axon_y[k], axon_z[k]),
                    axon_diam[k] / 2,
                    color=color,
                    fill=True,
                    alpha=alpha,
                )
            )
        if "extra_stim" in self:
            if self.extra_stim is not None:
                self.extra_stim.plot(axes=axes, color="gold", nerve_d=self.D)
        if num:
            for k in range(self.n_ax):
                axes.text(
                    self.axons_y[k],
                    self.axons_z[k],
                    str(k),
                    horizontalalignment="center",
                    verticalalignment="center",
                )
        axes.set_xlim(
            (
                -1.1 * self.D / 2 + self.y_grav_center,
                1.1 * self.D / 2 + self.y_grav_center,
            )
        )
        axes.set_ylim(
            (
                -1.1 * self.D / 2 + self.z_grav_center,
                1.1 * self.D / 2 + self.z_grav_center,
            )
        )

    def plot_block_summary(
        self,
        axes: plt.axes,
        AP_start: float,
        freq: float = None,
        t_refractory: float = 1,
        contour_color: str = "k",
        num: bool = False,
    ) -> None:
        """
        plot the block_summary of the fascicle in the Y-Z plane (transverse section)
        Color code:
        Green: fiber is blocked without any onset
        Blue: fiber is blocked with some onset
        Red: fiber is not blocked but has onset
        Grey: Fiber is nor blocked nor has onset

        A cross-mark on the fiber means block state can't be evaluted (is_blocked returned None)
        Alpha colorfill represents number of onset APs.

        Parameters
        ----------
        axes    : matplotlib.axes
            axes of the figure to display the fascicle
        AP_start : float
            timestamp of the test pulse start, in ms.
        freq : float, optional
            Frequency of the stimulation, for KES block, by default None
        t_refractory : float, optional
            Axon refractory period for myelinated fibers, by default 1
        contour_color   : str
            matplotlib color string applied to the contour. Black by default
        num             : bool
            if True, the index of each axon is displayed on top of the circle
        """

        ## plot contour
        axes.add_patch(
            plt.Circle(
                (self.y_grav_center, self.z_grav_center),
                self.D / 2,
                color=contour_color,
                fill=False,
                linewidth=2,
            )
        )
        ## plot axons
        axon_diam, _, axon_y, axon_z, is_blocked, n_onset = (
            self.get_block_summary_axons(
                AP_start=AP_start, freq=freq, t_refractory=t_refractory
            )
        )
        alpha_g = 1 / np.max(n_onset)

        # cmap = plt.get_cmap('viridis')
        # norm = plt.Normalize(min(n_onset), max(n_onset))
        # line_colors = cmap((n_onset))

        for k, _ in enumerate(axon_diam):
            if is_blocked[k] is not True:
                if n_onset[k] == 0:
                    c = "grey"
                    alpha = 0.5
                else:
                    c = "orangered"
                    alpha = n_onset[k] * alpha_g
                if is_blocked[k] is None:
                    axes.scatter(axon_y[k], axon_z[k], marker="x", s=20, c="k")

            else:
                if n_onset[k] == 0:
                    c = "seagreen"
                    alpha = 1
                else:
                    c = "steelblue"
                    alpha = n_onset[k] * alpha_g

            axes.add_patch(
                plt.Circle(
                    (axon_y[k], axon_z[k]),
                    axon_diam[k] / 2,
                    color=c,
                    fill=True,
                    alpha=alpha,
                )
            )

        if self.extra_stim is not None:
            self.extra_stim.plot(axes=axes, color="gold", nerve_d=self.D)
        if num:
            for k in range(self.n_ax):
                axes.text(
                    self.axons_y[k], self.axons_z[k], str(k)
                )  # horizontalalignment='center',verticalalignment='center')
        axes.set_xlim(
            (
                -1.1 * self.D / 2 + self.y_grav_center,
                1.1 * self.D / 2 + self.y_grav_center,
            )
        )
        axes.set_ylim(
            (
                -1.1 * self.D / 2 + self.z_grav_center,
                1.1 * self.D / 2 + self.z_grav_center,
            )
        )



""" NeuRon Virtualizer, large scale modeling of Peripheral Nervous System with random stimulation waveforms
optim utils submodule"""

from ._ContextModifiers import (
    context_modifier,
    stimulus_CM,
    biphasic_stimulus_CM,
    harmonic_stimulus_CM,
    harmonic_stimulus_with_pw_CM,
)
from ._CostEvaluation import (
    raster_count_CE,
    recrutement_count_CE,
    charge_quantity_CE,
    stim_energy_CE,
)
from ._OptimResults import optim_results
from ._OptimFunctions import (
    interpolate,
    interpolate_amp,
    interpolate_Npts,
    cost_position_saver,
)

submodules = []

classes = [
    "context_modifier",
    "stimulus_CM",
    "biphasic_stimulus_CM",
    "harmonic_stimulus_CM",
    "harmonic_stimulus_with_pw_CM",
    "raster_count_CE",
    "recrutement_count_CE",
    "charge_quantity_CE",
    "stim_energy_CE",
    "optim_results",
]

functions = [
    "interpolate",
    "interpolate_amp",
    "interpolate_Npts",
    "cost_position_saver",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



"""
log_interface handling.

For the moment, the log interface is managed using the pyswarms reporter class to avoid interference between loggers.
"""

import logging
import os
import sys
from rich.progress import track
from icecream import ic
from time import sleep

from ._parameters import parameters

from pyswarms.utils import Reporter


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def init_reporter():
    rep_nrv = Reporter(log_path=parameters.nrv_path + "/_misc/log/NRV.log")
    rep_nrv._default_config["handlers"]["default"]["level"] = "CRITICAL"
    rep_nrv._default_config["formatters"]["standard"][
        "format"
    ] = "%(asctime)s - %(message)s"
    rep_nrv._load_defaults()
    if os.path.isfile("report.log"):
        try:
            os.remove("report.log")
        except:
            pass
    return rep_nrv


def set_log_level(level, clear_log_file=False):
    if clear_log_file and os.path.isfile("report.log"):
        try:
            os.remove("report.log")
        except:
            pass
    if level is not None:
        rep_nrv._default_config["loggers"][""]["level"] = level
        rep_nrv._load_defaults()
    return rep_nrv


rep_nrv = init_reporter()


# !! ISSUE:
# Compatibility between logging and pyswarms when pyswarms is imported (not used necessarily)
# all messages log are also printed

# !! TEMPORARY SOLUTION
# Using a pyswarms reporter to log messages
# see old code commented

# logger = logging.getLogger()

# # logging config
# logging.basicConfig(
#     filename=dir_path + "/log/NRV.log",
#     level=logging.INFO,
#     format="%(asctime)s %(message)s",
#     datefmt="%m/%d/%Y %I:%M:%S %p",
# )


def rise_error(*args, out=1, **kwargs):
    """
    Rises and error to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing. This function exit the programm.

    Parameters
    ----------
    *args   :
        anything to pass as error message
    out     : int
        programm exit value, must be strictly positive. By default equal to 1
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]

    message = ""
    error = None
    for arg in args:
        print(type(arg), isinstance(arg, Exception))
        if isinstance(arg, type) and isinstance(arg(), Exception):
            error = arg
        else:
            message += str(arg)
    if parameters.LOG_Status:
        rep_nrv.log(message, lvl=logging.ERROR)
        # logger.error(message, lvl=logging.ERROR)
    if verbose and parameters.VERBOSITY_LEVEL >= 1:
        print(bcolors.FAIL + "NRV ERROR: " + message + bcolors.ENDC)
    else:
        err = (
            "NRV ERROR: "
            + message
            + "\n encountered in process "
            + parameters.proc_label
        )
        if parameters.LOG_Status:
            rep_nrv.log(err, lvl=logging.ERROR)
            # logger.error(err, lvl=logging.ERROR)
        if parameters.VERBOSITY_LEVEL >= 1:
            print(err)
            sys.stdout.flush()
    if out == 0:
        out = 1
    if not error is None:
        raise error(message)
    sys.exit(out)


def rise_warning(*args, abort=False, **kwargs):
    """
    Rises a warning to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing. This function can exit the programm.

    Parameters
    ----------
    *args   :
        anything to pass as warning message
    abort   : boolean
        if true, the programm exits with the value O (no error) and further computation are avoided,
        by default set to False.
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    message = ""
    for arg in args:
        message += str(arg)
    if parameters.is_alone:
        if parameters.LOG_Status:
            rep_nrv.log("NRV DEBUG: " + message, lvl=logging.DEBUG)
            # logger.warning("NRV WARNING: " + message)
        if verbose and parameters.VERBOSITY_LEVEL >= 2:
            print(bcolors.WARNING + "NRV WARNING: " + message + bcolors.ENDC)
    else:
        war = (
            "NRV WARNING: "
            + message
            + "\n encountered in process "
            + parameters.proc_label
        )
        if parameters.LOG_Status:
            rep_nrv.log(war, lvl=logging.DEBUG)
            # logger.warning(war)
        if parameters.VERBOSITY_LEVEL >= 2:
            print(bcolors.WARNING + war + bcolors.ENDC)
        elif parameters.VERBOSITY_LEVEL >= 4:
            print(bcolors.WARNING + war + bcolors.ENDC)
    if abort:
        sys.exit(0)


def pass_info(*args, **kwargs):
    """
    Pass an info to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing.

    Parameters
    ----------
    *args   :
        anything to pass as info
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    message = ""
    for arg in args:
        message += str(arg)
    if parameters.is_alone:
        if parameters.LOG_Status:
            rep_nrv.log("NRV DEBUG: " + message, lvl=logging.INFO)
            # logger.info("NRV DEBUG: " + message)
        if verbose and parameters.VERBOSITY_LEVEL >= 3:
            print("NRV INFO: " + message)
    else:
        inf = (
            "NRV INFO: "
            + message
            + "\n from process "
            + parameters.proc_label
        )
        if parameters.LOG_Status:
            rep_nrv.log(inf, lvl=logging.INFO)
            # logger.info(inf)
        if parameters.VERBOSITY_LEVEL >= 4:
            print(inf)
            sys.stdout.flush()


def pass_debug_info(*args, **kwargs):
    """
    Pass an info to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing.

    Parameters
    ----------
    *args   :
        anything to pass as info
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    message = ""
    for arg in args:
        message += str(arg)
    if parameters.is_alone:
        if parameters.LOG_Status and parameters.VERBOSITY_LEVEL >= 4:
            rep_nrv.log("NRV DEBUG: " + message, lvl=logging.DEBUG)
            # logger.info("NRV DEBUG: " + message)
        if verbose and parameters.VERBOSITY_LEVEL >= 4:
            print("NRV DEBUG: " + message)
    else:
        inf = (
            "NRV DEBUG: "
            + message
            + "\n from process "
            + parameters.proc_label
        )
        if parameters.LOG_Status:
            rep_nrv.log(inf, lvl=logging.DEBUG)
            # logger.info(inf)
        if parameters.VERBOSITY_LEVEL >= 4:
            print(inf)
            sys.stdout.flush()


def progression_popup(current, max_iter, begin_message="", end_message="", endl=""):
    """
    Displays the progression on prompt, for single process only, nothing saved to the log

    Parameters
    ----------
    current         : int
        current iteration
    max_iter        : int
        maximum iteration number
    begin_message   : str
        additional message placed at the start of the progression line
    end_message     : str
        additional message placed after iteration and max at the end of the progression line
    endl            : str
        line termination, by default empty string
    """
    if parameters.is_alone:
        print(
            begin_message + f"{current + 1}" + "/" + str(max_iter) + end_message,
            end=endl,
        )
        if current == max_iter - 1:
            print("\n")


def prompt_debug(*args):
    """
    outputs for debug, simple redirection to icecream without the installation (see icecream help).
    Could be interesting to redirect also to the log...

    Parameters
    ----------
    *args   :
        anything to pass to icecrean
    """
    ic(*args)


def clear_prompt_line(n: int = 1) -> None:
    """
    Clear lines of the prompt, to overwrite stuffs.

    Parameters
    ----------
    n : int, optional
        Number of line to clear, by default 1
    """
    LINE_UP = "\033[1A"
    LINE_CLEAR = "\x1b[2K"
    for i in range(n):
        print(LINE_UP, end=LINE_CLEAR)



from ..nmod._nerve import nerve
from ..nmod._fascicles import fascicle
from ..nmod._myelinated import get_MRG_parameters


def crop_fascicle(fasc: fascicle, x0: float, new_l: float) -> fascicle:
    """_summary_

    Parameters
    ----------
    fasc : fascicle
        fascicle to crop
    x0 : float
        x position of the begining of the new fascicle
    new_l : float
        length of the new fascicle

    Returns
    -------
    fascicle
        fascicle cropped
    """
    fasc.define_length(new_l)
    if not (fasc.L is None or len(fasc.NoR_relative_position) == 0):
        for i in range(fasc.n_ax):
            if fasc.axons_type[i] == 1:
                _, _, _, _, _, deltax, _, _ = get_MRG_parameters(fasc.axons_diameter[i])
                fasc.NoR_relative_position = fasc.NoR_relative_position + (x0 / deltax)

                fasc.NoR_relative_position -= int(fasc.NoR_relative_position)
    return fasc


def crop_nerve(nerv: nerve, x0: float, new_l: float) -> nerve:
    """_summary_

    Parameters
    ----------
    nerv : nerve
        nerve to crop
    x0 : float
        x position of the begining of the new fascicle
    new_l : float
        length of the new fascicle

    Returns
    -------
    nerve
        nerve cropped
    """
    nerv.define_length(new_l)
    for fasc in nerv.fascicles.values:
        fasc = crop_fascicle(fasc, x0, new_l)
    return nerv



""" Utils: general purpose functions and classes

utils provides some classes and functions for runing and interconnecting code
from the scientific sub-packages.
"""

from ._misc import (
    distance_point2point,
    distance_point2line,
    nearest_idx,
    nearest_greater_idx,
    in_tol,
    get_perineurial_thickness,
    membrane_capacitance_from_model,
    compute_complex_admitance,
    get_MRG_parameters,
    get_length_from_nodes,
)
from ._nrv_function import (
    nrv_function,
    function_1D,
    gaussian,
    gate,
    function_2D,
    ackley,
    beale,
    goldstein_price,
    booth,
    bukin6,
    function_ND,
    Id,
    rosenbock,
    rastrigin,
    sphere,
    cost_evaluation,
    nrv_interp,
    MeshCallBack,
)

from ._units import (
    print_default_nrv_unit,
    from_nrv_unit,
    to_nrv_unit,
    convert,
    sci_round,
)

from ._stimulus import (
    is_stim,
    set_common_time_series,
    get_equal_timing_copies,
    datfile_2_stim,
    stimulus,
)


submodules = []

classes = [
    "nrv_function",
    "function_1D",
    "gaussian",
    "gate",
    "function_2D",
    "ackley",
    "beale",
    "goldstein_price",
    "booth",
    "bukin6",
    "function_ND",
    "Id",
    "rosenbock",
    "rastrigin",
    "sphere",
    "cost_evaluation",
    "nrv_interp",
    "MeshCallBack",
    "stimulus",
]

functions = [
    "distance_point2point",
    "distance_point2line",
    "nearest_idx",
    "nearest_greater_idx",
    "in_tol",
    "get_perineurial_thickness",
    "membrane_capacitance_from_model",
    "compute_complex_admitance",
    "print_default_nrv_unit",
    "from_nrv_unit",
    "to_nrv_unit",
    "convert",
    "sci_round",
    "is_stim",
    "set_common_time_series",
    "get_equal_timing_copies",
    "datfile_2_stim",
    "get_MRG_parameters",
    "get_length_from_nodes",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



"""
NRV-:class:`.nerve_results` handling.
"""

from ...backend._NRV_Results import sim_results
from ._fascicles_results import fascicle_results
from ._axons_results import axon_results
from ...backend._log_interface import rise_error, rise_warning, pass_info
from ...utils._units import nm, convert, from_nrv_unit
from ...utils._misc import membrane_capacitance_from_model


import matplotlib.pyplot as plt
import numpy as np


def number_in_str(s: str) -> bool:
    return any(i.isdigit() for i in s)


class nerve_results(sim_results):
    """ """

    def __init__(self, context=None):
        super().__init__(context)

    @property
    def fascicle_keys(self) -> list:
        """
        Number of fascicles in the fascicle
        """
        all_keys = self.keys()
        return [i for i in all_keys if ("fascicle" in i and number_in_str(i))]

    @property
    def n_fasc(self) -> int:
        """
        Number of fascicles in the fascicle
        """
        return len(self.fascicles)

    @property
    def n_ax(self) -> int:
        """
        Number of axons in the fascicle
        """
        fasc_keys = self.fascicle_keys
        _n_ax = 0
        for key in fasc_keys:
            _n_ax += self[key].n_ax
        return _n_ax

    @property
    def axons_type(self) -> np.ndarray:
        """
        type of axons of each fascicles

        Returns
        -------
        np.ndarray (self.n_ax, 2)
            _description_
        """
        return self.axons_pop_properties[:,[0,2]]

    @property
    def fasc_properties(self) -> np.ndarray:
        """
        Porperties of axons population of each fascicles

        Returns
        -------
        np.ndarray (self.n_ax, 6)
            ndarray gathering, for all fascicles in the nerve, fascicle IDs, diameter and y and z positions.
        """
        fasc_keys = self.fascicle_keys
        _fasc = np.zeros((self.n_fasc, 4))
        for i_fasc, key in enumerate(fasc_keys):
            fasc_ = self[key]
            _fasc[i_fasc, :] = np.array([
                fasc_.ID,
                fasc_.D,
                fasc_.y_grav_center,
                fasc_.z_grav_center,
            ])
        return _fasc

    @property
    def axons_pop_properties(self) -> np.ndarray:
        """
        Porperties of axons population of each fascicles

        Returns
        -------
        np.ndarray (self.n_ax, 6)
            ndarray gathering, for all axons in the nerve, corresponding fascicle and axon IDs, myelinating type, diameter and y and z positions.
        """
        fasc_keys = self.fascicle_keys
        _mye = np.zeros((self.n_ax, 6))
        _offset = 0
        for key in fasc_keys:
            fasc_n_ax = self[key].n_ax
            fasc_axons = np.vstack(
                (
                    self[key].ID * np.ones(fasc_n_ax),
                    np.arange(fasc_n_ax),
                    self[key].axons_type,
                    self[key].axons_diameter,
                    self[key].axons_y,
                    self[key].axons_z,
                )
            ).T
            _mye[_offset : _offset + fasc_n_ax, :] = fasc_axons
            _offset += fasc_n_ax
        return _mye

    def get_fascicle_results(self, ID: int) -> fascicle_results:
        if ID not in self.fascicles_IDs:
            rise_error(("Fascicle ID does not exists."))
        else:
            return self[f"fascicle{ID}"]

    def get_axon_results(self, fasc_ID: int, ax_ID: int) -> axon_results:
        fasc_ID = int(fasc_ID)
        ax_ID = int(ax_ID)
        if fasc_ID not in self.fascicles_IDs:
            rise_error(f"Fascicle ID: {fasc_ID} does not exists.")
        else:
            if ax_ID >= self[f"fascicle{fasc_ID}"].n_ax:
                rise_error(f"Axon ID: {ax_ID} does not exists in Fascicle {fasc_ID} .")
            else:
                return self[f"fascicle{fasc_ID}"][f"axon{ax_ID}"]

    def get_fascicle_key(self) -> list:
        all_keys = self.keys()
        fascicle_keys = [i for i in all_keys if ("fascicle" in i and number_in_str(i))]
        return fascicle_keys

    def get_recruited_axons(
        self, ax_type: str = "all", normalize: bool = False
    ) -> int | float:
        """
        Return the number or the ratio of recruited axons in the nerve

        Parameters
        ----------
        diam : float
            diameter above wich axon should be counted.
        ax_type : str, optional
            type of axon counted,possible options:
             - "all" (default)
             - "unmyelinated"
             - "myelinated"
        normalize : bool, optional
            if False the total number of recruited axons is returned, else the ratio is returned, by default False

        Returns
        -------
        float
            number of recruited axons
        """
        fasc_keys = self.fascicle_keys
        n_recr = 0
        for key in fasc_keys:
            fasc_res = self[key]
            n_recr += fasc_res.get_recruited_axons(ax_type=ax_type, normalize=normalize)
        if normalize:
            n_recr /= self.n_fasc
        return n_recr

    # impeddance related methods
    def get_membrane_conductivity(
        self, x: float = 0, t: float = 0, unit: str = "S/cm**2", mem_th: float = 7 * nm
    ) -> np.array:
        """
        get the membran conductivity of each axon at a position x and a time t

        Parameters
        ----------
        x : float, optional
            x-position in um where to get the conductivity, by default 0
        t : float, optional
            simulation time in ms when to get the conductivity, by default 0
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm

        Note
        ----
        depending of the unit parameter this function either return :

            - the surface conductivity in [S]/([m]*[m]): from neuron simulation
            - the conductivity in [S]/[m]:  by multiplying surface conductivity by membrane thickness
        """

        g = []
        f_keys = self.get_fascicle_key()
        for key in f_keys:
            g_ = self[key].get_membrane_conductivity(x=x, t=t, unit=unit, mem_th=mem_th)
            if g_ is not None:
                g = np.concatenate((g, g_))
            else:
                return None
        return g

    def get_membrane_capacitance(
        self, unit: str = "uF/cm**2", mem_th: float = 7 * nm
    ) -> tuple[float]:
        """
        get the membrane capacitance or permitivity of unmyelinated and myelinated axons filling the ner

        Parameters
        ----------
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm

        Note
        ----
        depending of the unit parameter this function either return :

            - the surface conductivity in [S]/([m]*[m]): from neuron simulation
            - the conductivity in [S]/[m]:  by multiplying surface conductivity by membrane thickness
        """
        u_c_mem = membrane_capacitance_from_model(self.unmyelinated_param["model"])
        m_c_mem = membrane_capacitance_from_model(self.myelinated_param["model"])

        # Surface capacity in [F]/([m]*[m])
        if "2" in unit:
            return convert(u_c_mem, "S/cm**2", unit), convert(m_c_mem, "S/cm**2", unit)
        # permitivity in [F]/[m]
        else:
            u_c_mem *= from_nrv_unit(mem_th, "cm")
            m_c_mem *= from_nrv_unit(mem_th, "cm")
            return convert(u_c_mem, "S/cm", unit), convert(m_c_mem, "S/cm", unit)

    def get_membrane_complexe_admitance(
        self,
        f: float = 1.0,
        x: float = 0,
        t: float = 0,
        unit: str = "S/m",
        mem_th: float = 7 * nm,
    ) -> np.array:
        """
        get the membran complexe admitance of each axon at a position x and a time t for a given frequency

        Parameters
        ----------
        f : float or np.array, optional
            effective frequency in kHz, by default 1
        x : float, optional
            x-position in um where to get the conductivity, by default 0
        t : float, optional
            simulation time in ms when to get the conductivity, by default 0
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm
        """
        Y = []
        f_keys = self.get_fascicle_key()
        for key in f_keys:
            Y_ = self[key].get_membrane_complexe_admitance(
                f=f, x=x, t=t, unit=unit, mem_th=mem_th
            )
            if Y_ is not None:
                Y = np.concatenate((Y, Y_))
            else:
                return None
        return Y

    ## Representation methods
    def plot_recruited_fibers(
        self,
        axes: plt.axes,
        contour_color: str = "k",
        myel_color: str = "r",
        unmyel_color: str = "b",
        elec_color: str = "gold",
        num: bool = False,
        **kwgs,
    ):
        """ """
        ## plot contour
        axes.add_patch(
            plt.Circle(
                (self.y_grav_center, self.z_grav_center),
                self.D / 2,
                color=contour_color,
                fill=False,
                linewidth=4,
            )
        )
        fasc_keys = self.get_fascicle_key()
        for key in fasc_keys:
            fasc_res = self[key]
            fasc_res.plot_recruited_fibers(
                axes=axes,
                contour_color="grey",
                myel_color=myel_color,
                unmyel_color=unmyel_color,
                num=num,
            )
        
        if "extra_stim" in self:
            if self.extra_stim is not None:
                self.extra_stim.plot(axes=axes, color=elec_color, nerve_d=self.D)
        axes.set_xlim((-1.1 * self.D / 2, 1.1 * self.D / 2))
        axes.set_ylim((-1.1 * self.D / 2, 1.1 * self.D / 2))



"""
NRV-Cellular Level simulations.
"""
import sys
from typing import Callable
from time import perf_counter
from multiprocessing import Pool
# from pathos.multiprocessing import ProcessingPool as Pool
from rich.progress import Progress, SpinnerColumn, BarColumn, TextColumn, TimeRemainingColumn
from functools import partial
from itertools import repeat

from ..backend._file_handler import is_iterable
from ..backend._log_interface import pass_info, rise_error, rise_warning, clear_prompt_line
from ..backend._parameters import parameters
from ..fmod._electrodes import *
from ..fmod._extracellular import *
from ..fmod._materials import *
from ..nmod._axons import *
from ..nmod._myelinated import *
from ..nmod._unmyelinated import *
from ..utils._stimulus import *
from ..utils._saving_handler import *
from ._axon_postprocessing import *


def set_args_kwargs(func, args, kwargs):   
    return func(*args, **kwargs)

def _call_wrapper(args):
    param, kw, func = args
    return func(param, **kw)

def search_threshold_dispatcher(search_func: Callable, parameter_list: list[any], ncore: int=None, **kwargs)->list[any]:
    """
    Automatically dispatches any search threshold callable object on available cpu cores. 

    Parameters
    ----------
    search_func : Callable
        Any callable object that takes as input a value from parameter_list and that returns a threshold
    parameter_list : list[any]
        Lists the values to evaluate the thresholds
    ncore : int, optional
        Fix the number of CPU core count. If None, ncore is set the size of parameter_list. By default None

    Returns
    -------
    list[any]
        List of ordered thresholds.
    """
    
    total = len(parameter_list)
    if ncore is None:
        ncore = total

    # Split constant kwargs variable ones
    varying_keys = {k: v for k, v in kwargs.items() if is_iterable(v) and len(v) == total}
    constant_kwargs = {k: v for k, v in kwargs.items() if not (is_iterable(v) and len(v) == total)}

    call_args = []
    for i in range(total):
        kw = {k: v[i] for k, v in varying_keys.items()}
        kw.update(constant_kwargs)
        call_args.append((parameter_list[i], kw, search_func)) 

    call_args = []
    for i in range(total):
        kw = {k: v[i] for k, v in varying_keys.items()}
        kw.update(constant_kwargs)
        call_args.append((parameter_list[i], kw, search_func))

    results = []
    #TODO: display individual search output (for each core, see EIT) instead of "processing search"
    with Pool(processes=ncore) as pool:
        with Progress(
            SpinnerColumn(),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TimeRemainingColumn(),
            TextColumn("Processing Search..."),
        ) as progress:
            task = progress.add_task("[cyan]Dispatching...", total=total)
            for result in pool.imap(_call_wrapper, call_args):
                results.append(result)
                progress.update(task, advance=1,refresh=True)

    return results



def axon_AP_threshold(axon: axon, amp_max: float, update_func: Callable,
                      t_sim: float = 5, tol: float = 1, t_start = None, freq = None,
                      verbose: bool = True, args_update = None, save_path: str|None = None, **kwargs) -> np.float64 :
    """
    Find the activation threshold of an axon with arbitrary stimulation settings.

    Parameters
    ----------
    axon : axon
        axon object to simulation
    amp_max : float
        maximum amplitude for the binary search, in µA
    update_func : Callable
        Callable function to update the axon stimulation parameters between each binary search iteration
    t_sim : float, optional
        Simulation duration, in ms, by default 5
    tol : float, optional
        Search tolerance, in %, by default 1
    verbose : bool, optional
        Verbosity of the search, by default True
    args_update : arg, optional
        update_func arguments, by default None
    save_path : str, optional
        if not none, raster_plot and plot_x_t of axon results are saved in save_path

    Returns
    -------
    current_amp : np.float64
        The threshold found with the binary search.
    """

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0
    
    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = tol

    vm_key = "V_mem"
    if (freq is not None):
        vm_key += "_filtered"

    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    t_start_cnt = perf_counter()

    while keep_going:
        if verbose and Niter == 1:
            pass_info(f"Iteration {Niter}, Amp is {np.round(current_amp,2)}µA ...")
            
        update_func(axon,current_amp,**args_update)
        verb = parameters.get_nrv_verbosity()
        parameters.set_nrv_verbosity(2)
        results = axon.simulate(t_sim=t_sim)
        parameters.set_nrv_verbosity(verb)

        if freq is not None:
            results.filter_freq("V_mem", freq, Q=2)

        if save_path:
            fig, ax = plt.subplots(1)
            results.raster_plot(ax,key = vm_key)
            ax.set_title(f"Amplitude: {np.round(current_amp,2)}µA")
            amp_str = str(np.round(current_amp,2)).replace(".", "_")
            fig.tight_layout()
            fig_name = save_path + f"raster_plot_axon_{axon.ID}_amp_{amp_str}_uA_iter_{Niter}.png"
            fig.savefig(fig_name)
            plt.close(fig)

            fig, ax = plt.subplots(1)
            results.plot_x_t(ax,key = vm_key)
            ax.set_title(f"Amplitude: {np.round(current_amp,2)}µA")
            fig.tight_layout()
            fig_name = save_path + f"vmen_plot_axon_{axon.ID}_amp_{amp_str}_uA_iter_{Niter}.png"
            fig.savefig(fig_name)
            plt.close(fig)

        if verbose:
            clear_prompt_line(1)

        # post-process results
        delta_amp = np.abs(current_amp - previous_amp)
        #put in a function
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                current_tol = 100 * delta_amp / current_amp
            else:
                current_tol = 100 * delta_amp / previous_amp
            if current_tol <= tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if results.is_recruited(vm_key=vm_key,t_start=t_start):
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info(f"Iteration {Niter}, Amp is {np.round(current_amp,2)}µA"+ 
                          f" ({np.round(current_tol,2)}%)... AP Detected! (in {np.round(results["sim_time"],3)}s)")
                #pass_info("... Spike triggered")
            amplitude_max_th = previous_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info(f"Iteration {Niter}, Amp is {np.round(current_amp,2)}µA"+
                          f" ({np.round(current_tol,2)}%)... AP Not Detected! (in {np.round(results["sim_time"],3)}s)")
            current_amp = amplitude_max_th - delta_amp / 2
            amplitude_min_th = previous_amp

        if previous_amp == amp_max:
            current_amp = amp_min

        Niter += 1
    t_stop = perf_counter()
    if verbose:
        clear_prompt_line(1)
        pass_info(f"Activation threshold is {np.round(current_amp,2)}µA ({np.round(current_tol,2)}%),"
                  + f" found in {Niter-1} iterations ({np.round(t_stop-t_start_cnt,2)}s).")
    return current_amp

def axon_block_threshold(axon: axon, amp_max: float, update_func: Callable, AP_start: float,
                      t_sim: float = 10, tol: float = 1, freq: float|None = None, verbose: bool = True, 
                      args_update = dict|None, save_path: str|None = None, **kwargs) -> np.float64 :
    """
    Find the block threshold of an axon with arbitrary stimulation settings.

    Parameters
    ----------
    axon : axon
        axon object to simulation
    amp_max : float
        maximum amplitude for the binary search, in µA
    update_func : Callable
        Callable object to update the axon stimulation parameters between each binary search iteration
    AP_start : float
            timestamp of the test pulse start, in ms.
    t_sim : float, optional
        Simulation duration, in ms, by default 5
    tol : float, optional
        Search tolerance, in %, by default 1
    freq : float, optional
            Frequency of the stimulation, for KES block, by default None
    verbose : bool, optional
        Verbosity of the search, by default True
    args_update : arg, optional
        update_func arguments, by default None
    save_path : str, optional
        if not none, raster_plot and plot_x_t of axon results are saved in save_path

    Returns
    -------
    current_amp : np.float64
        The block threshold found with the binary search.
    """

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0
    
    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = tol

    vm_key = "V_mem"
    if (freq is not None):
        vm_key += "_filtered"

    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    t_start_cnt = perf_counter()
    
    while keep_going:
        if verbose and Niter == 1:
            pass_info(f"Iteration {Niter}, Amp is {np.round(current_amp,2)}µA ...")
            
        update_func(axon,current_amp,**args_update)

        verb = parameters.get_nrv_verbosity()
        #parameters.set_nrv_verbosity(2)
        results = axon.simulate(t_sim=t_sim)
        #parameters.set_nrv_verbosity(verb)

        is_blocked = results.is_blocked(AP_start=AP_start, freq=freq)
        if save_path:
            fig, ax = plt.subplots(1)
            results.raster_plot(ax,key = vm_key)
            ax.set_title(f"Block Amplitude: {np.round(current_amp,2)}µA")
            amp_str = str(np.round(current_amp,2)).replace(".", "_")
            fig.tight_layout()
            fig_name = save_path + f"raster_plot_axon_{axon.ID}_amp_{amp_str}_uA_iter_{Niter}.png"
            fig.savefig(fig_name)
            plt.close(fig)

            fig, ax = plt.subplots(1)
            results.plot_x_t(ax,key = vm_key)
            ax.set_title(f"Block Amplitude: {np.round(current_amp,2)}µA")
            fig.tight_layout()
            fig_name = save_path + f"vmen_plot_axon_{axon.ID}_amp_{amp_str}_uA_iter_{Niter}.png"
            fig.savefig(fig_name)
            plt.close(fig)

        if verbose:
            clear_prompt_line(1)

        # post-process results
        delta_amp = np.abs(current_amp - previous_amp)
        #put in a function
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                current_tol = 100 * delta_amp / current_amp
            else:
                current_tol = 100 * delta_amp / previous_amp
            if current_tol <= tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if is_blocked is None:
            rise_warning("Failed to evaluate block state of the axon! Consider changing the simulation parameters.")
            return 0.0
        
        if is_blocked:
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0.0
            if verbose:
                pass_info(f"Iteration {Niter}, Amp is {np.round(current_amp,2)}µA"+ 
                          f" ({np.round(current_tol,2)}%)... AP Blocked! (in {np.round(results["sim_time"],3)}s)")
                #pass_info("... Spike triggered")
            amplitude_max_th = previous_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0.0
            if verbose:
                pass_info(f"Iteration {Niter}, Amp is {np.round(current_amp,2)}µA"+
                          f" ({np.round(current_tol,2)}%)... AP Not Blocked! (in {np.round(results["sim_time"],3)}s)")
            current_amp = amplitude_max_th - delta_amp / 2
            amplitude_min_th = previous_amp

        if previous_amp == amp_max:
            current_amp = amp_min

        Niter += 1
    t_stop = perf_counter()
    if verbose:
        clear_prompt_line(1)
        pass_info(f"Block threshold is {np.round(current_amp,2)}µA ({np.round(current_tol,2)}%),"
                  + f" found in {Niter-1} iterations ({np.round(t_stop-t_start_cnt,2)}s).")
    return current_amp

def firing_threshold_point_source(
    diameter,
    L,
    dist_elec,
    cath_time=100e-3,
    model="MRG",
    amp_max=2000,
    amp_tol=1,
    dt = 0.005,
    verbose=True,
    **kwargs,
):
    """
    Find the firing threshold with point source approximation using binary search.

    Parameters
    ----------
    diameter        : float
            axon diameter in um
    L               : float
            axon length in um
    dist_elec       : float
            y coordinate of the point source electrode in um
    cath_time      : float
            cathodic pulse width in ms
    model           : str
            Axon model
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    dt              : float
        Set the dt to a specific value in ms
    verbose         : bool
            set the verbosity of the search
    **kwargs:
            See below

    Keyword Arguments:
            material       :   str
                    material used to compute the extracellular field
            position_elect : float
                    x location of the electrode, between 0 and 1
            z_elect        : float
                    z coordinate of the electrode
            amp_min        : float
                    minimum amplitude for the binary search
            f_dlambda      : float
                    To set the number of segment with the d_dlambda rule (Hz)
            t_sim          : float
                    Duration of the simulation in ms
            amp_tol_abs    : float
                    Specify an absolute amplitude tolerance for the binary search
                    in uA
            anod_first  : bool
                    Set the anodic phase before the cathodic phase
            t_inter : float
                    Specify the interphase delay in ms
            cath_an_ratio: float
                    Specify the cathodic/anodic ratio
            dt              : float
                    Set the dt to a specific value in ms
            cath_an_ratio          : float
                    Set the tolorance for dt in %
            nseg            : int
                    Set the number of segment for the axon
            amp_tol_abs     : float
                    Set the absolute tolerance for the binary search in uA.
            node_shift      : float between -1 and 1
                    Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
            n_nodes         : integer
                    Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

    Returns
    -------
    threshold       : float
            estimated firing threshold in uA
    """

    rise_warning(
        "DeprecationWarning: ",
        "firing_threshold_point_source is obsolete use axon_AP_threshold instead"
    )

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        if model in myelinated_models:
            t_sim = 5
        else:
            t_sim = 20

    if "material" in kwargs:
        material = kwargs.get("material")
    else:
        material = material = "endoneurium_bhadra"

    if "anod_first" in kwargs:
        anod_first = kwargs.get("anod_first")
    else:
        anod_first = False

    if "t_inter" in kwargs:
        t_inter = kwargs.get("t_inter")
    else:
        t_inter = 0

    if "cath_an_ratio" in kwargs:
        cath_an_ratio = kwargs.get("cath_an_ratio")
    else:
        cath_an_ratio = 1

    if "position_elec" in kwargs:
        position_elec = kwargs.get("position_elec")
    else:
        position_elec = 0.5

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "f_dlambda" in kwargs:
        f_dlambda = kwargs.get("f_dlambda")
    else:
        f_dlambda = 0

    if "nseg" in kwargs:
        nseg = kwargs.get("nseg")
    else:
        nseg = 0

    if "Nseg_per_sec" in kwargs:
        Nseg_per_sec = kwargs.get("Nseg_per_sec")
    else:
        Nseg_per_sec = 0

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "node_shift" in kwargs:
        node_shift = kwargs.get("node_shift")
    else:
        node_shift = 0

    if "n_nodes" in kwargs:
        n_nodes = kwargs.get("n_nodes")
    else:
        n_nodes = 0

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol
    # axon
    y = 0
    z = 0

    # extra cellular
    extra_material = load_material(material)
    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing firing current amplitude "
                + str(current_amp)
                + " uA"
            )
        # create axon
        if model in unmyelinated_models:
            if Nseg_per_sec:
                axon1 = unmyelinated(
                    y, z, diameter, L, dt=dt, model=model, Nseg_per_sec=Nseg_per_sec
                )  # freq=f_dlambda
            else:
                axon1 = unmyelinated(
                    y, z, diameter, L, dt=dt, freq=f_dlambda, model=model
                )
        elif model in myelinated_models:
            if n_nodes > 0:
                L = get_length_from_nodes(diameter, n_nodes)
            if Nseg_per_sec:
                axon1 = myelinated(
                    y,
                    z,
                    diameter,
                    L,
                    rec="nodes",
                    dt=dt,
                    model=model,
                    Nseg_per_sec=Nseg_per_sec,
                )  # freq=f_dlambda
            else:
                axon1 = myelinated(
                    y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
                )
        else:
            rise_error("Error: Specified model is not recognized.")
        # extra-cellular stimulation
        x_elec = L * position_elec
        if model in myelinated_models:  # Align electrode with node of Ranvier
            n_nodes = len(axon1.x_nodes)
            nearest_node = np.int32(np.round(position_elec * n_nodes))
            x_elec = axon1.x_nodes[nearest_node]
            if node_shift != 0:  # shift it if needed
                if node_shift > 0:
                    if nearest_node < n_nodes - 1:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec + d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )
                else:
                    if nearest_node != 0:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec - d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )

        y_elec = dist_elec
        z_elec = 0
        elec_1 = point_source_electrode(x_elec, y_elec, z_elec)

        # stimulus def
        stim_1 = stimulus()
        start = 1
        I_cathod = current_amp
        if cath_an_ratio > 0:
            I_anod = I_cathod / cath_an_ratio

            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, I_anod, t_inter, anod_first=anod_first
            )
        else:
            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, 0, 0, anod_first=anod_first
            )

        stim_extra = stimulation(extra_material)
        stim_extra.add_electrode(elec_1, stim_1)
        axon1.attach_extracellular_stimulation(stim_extra)
        # simulate axon activity
        results = axon1.simulate(t_sim=t_sim)
        del axon1
        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if results.is_recruited("V_mem") :
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike triggered")
            amplitude_max_th = previous_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not triggered")
            current_amp = amplitude_max_th - delta_amp / 2
            amplitude_min_th = previous_amp

        if previous_amp == amp_max:
            current_amp = amp_min

        Niter += 1
    return current_amp


def firing_threshold_from_axon(
    axon, cath_time=100e-3, amp_max=2000, amp_tol=1, verbose=True, **kwargs
):
    """
    Find the firing threshold from a specified axon using binary search

    Parameters
    ----------
    axon        : axon class
            axon class
    cath_time      : float
            cathodic pulse width in ms
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    verbose         : bool
            set the verbosity of the search
    **kwargs:
            See below

    Keyword Arguments:
    elec_id        : int
            elec_id where to change stimulus
    amp_min        : float
            minimum amplitude for the binary search
    t_sim          : float
            Duration of the simulation in ms
    amp_tol_abs    : float
            Specify an absolute amplitude tolerance for the binary search
            in uA
    anod_first  : bool
            Set the anodic phase before the cathodic phase
    t_inter : float
            Specify the interphase delay in ms
    cath_an_ratio: float
            Specify the cathodic/anodic ratio
    cath_an_ratio          : float
            Set the tolorance for dt in %
    amp_tol_abs     : float
            Set the absolute tolerance for the binary search in uA.

    Returns
    -------
    threshold       : float
            estimated firing threshold in uA
    """

    rise_warning(
        "DeprecationWarning: ",
        "firing_threshold_point_source is obsolete use axon_AP_threshold instead"
    )

    if "elec_id" in kwargs:
        elec_id = kwargs.get("elec_id")
    else:
        elec_id = 0

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        t_sim = 10

    if "anod_first" in kwargs:
        anod_first = kwargs.get("anod_first")
    else:
        anod_first = False

    if "t_inter" in kwargs:
        t_inter = kwargs.get("t_inter")
    else:
        t_inter = 0

    if "cath_an_ratio" in kwargs:
        cath_an_ratio = kwargs.get("cath_an_ratio")
    else:
        cath_an_ratio = 1

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol

    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1

    # axon.save_axon(save=True, fname='_axon.json', extracel_context=True)
    # del (axon)
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing firing current amplitude "
                + str(current_amp)
                + " uA"
            )

        # stimulus def
        stim_1 = stimulus()
        start = 1
        I_cathod = current_amp
        # I_cathod =
        if cath_an_ratio > 0:
            I_anod = I_cathod / cath_an_ratio
            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, I_anod, t_inter, anod_first=anod_first
            )
        else:
            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, 0, 0, anod_first=anod_first
            )
        # axon_load = load_any_axon(axon, extracel_context=True)
        # axon_th = copy.deepcopy(axon)
        axon.change_stimulus_from_electrode(elec_id, stim_1)

        # simulate axon activity
        results = axon.simulate(t_sim=t_sim, loaded_footprints=True)
        # del (axon)
        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if results.is_recruited("V_mem"):
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike triggered")
            amplitude_max_th = previous_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not triggered")
            current_amp = amplitude_max_th - delta_amp / 2
            amplitude_min_th = previous_amp

        if previous_amp == amp_max:
            current_amp = amp_min

        Niter += 1
    del axon
    return current_amp


def blocking_threshold_point_source(
    diameter,
    L,
    dist_elec,
    block_freq,
    model="MRG",
    amp_max=2000,
    amp_tol=1,
    dt=0.005,
    Nseg_per_sec=2,
    verbose=True,
    **kwargs,
):
    """
    Find the blocking threshold with point source approximation using binary search.

    Parameters
    ----------
    diameter        : float
            axon diameter in um
    L               : float
            axon length in um
    dist_elec       : float
            y coordinate of the point source electrode in um
    block_freq      : float
            Blocking frequency in kHz
    model           : str
            Axon model
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    dt              : float
            time discretization in ms
    Nseg_per_sec    : int
            Number of segment per section (myelinated axons)
            Number of segment per mm of length (unmyelinated axons)
    verbose         : bool
            set the verbosity of the search

    **kwargs:
            See below

    Keyword Arguments:
            material       :   str
                    material used to compute the extracellular field
            position_elect : float
                    x location of the electrode, between 0 and 1
            z_elect        : float
                    z coordinate of the electrode
            amp_min        : float
                    minimum amplitude for the binary search
            f_dlambda      : float
                    To set the number of segment with the d_dlambda rule (Hz)
            t_sim          : float
                    Duration of the simulation in ms
            amp_tol_abs    : float
                    Specify an absolute amplitude tolerance for the binary search
                    in uA
            t_position     : float
                    Positition on the test electrode.
                    Relative position for unmyelinated axons.
                    Node number for myelinated axons.
            t_start        : float
                    start of the test pulse, in ms
            t_duration     : float
                    duration of the test pulse, in ms
            t_amplitude    : float
                    Amplitude of the test pulse, in nA
            b_start        : float
                    start of the blocking stimulation in ms
            b_duration     : float
                    duration of the blocking stimulation in ms
            node_shift      : float between -1 and 1
                    Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
            n_nodes         : integer
                    Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

    Returns
    -------
    threshold       : float
            estimated threshold in uA
    """

    rise_warning(
        "DeprecationWarning: ",
        "blocking_threshold_point_source is obsolete use axon_block_threshold instead"
    )


    if "material" in kwargs:
        material = kwargs.get("material")
    else:
        material = "endoneurium_bhadra"

    if "position_elec" in kwargs:
        position_elec = kwargs.get("position_elec")
    else:
        position_elec = 0.5

    if "z_elec" in kwargs:
        z_elec = kwargs.get("z_elec")
    else:
        z_elec = 0

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "f_dlambda" in kwargs:
        f_dlambda = kwargs.get("f_dlambda")
    else:
        f_dlambda = 100

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        t_sim = 40

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "t_position" in kwargs:
        t_position = kwargs.get("t_position")
    else:
        t_position = 0.05

    if "t_start" in kwargs:
        t_start = kwargs.get("t_start")
    else:
        t_start = 30

    if "t_duration" in kwargs:
        t_duration = kwargs.get("t_duration")
    else:
        t_duration = 1

    if "t_amplitude" in kwargs:
        t_amplitude = kwargs.get("t_amplitude")
    else:
        t_amplitude = 2

    if "b_start" in kwargs:
        b_start = kwargs.get("b_start")
    else:
        b_start = 3

    if "b_duration" in kwargs:
        b_duration = kwargs.get("b_duration")
    else:
        b_duration = t_sim

    if "node_shift" in kwargs:
        node_shift = kwargs.get("node_shift")
    else:
        node_shift = 0

    if "n_nodes" in kwargs:
        n_nodes = kwargs.get("n_nodes")
    else:
        n_nodes = 0

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol

    y = 0
    z = 0

    extra_material = load_material(material)
    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing block current amplitude "
                + str(current_amp)
                + " uA"
            )
        # create axon
        if model in unmyelinated_models:
            if Nseg_per_sec:
                axon1 = unmyelinated(
                    y,
                    z,
                    diameter,
                    L,
                    dt=dt,
                    Nseg_per_sec=Nseg_per_sec,
                    Nsec=1,
                    model=model,
                )
            else:
                axon1 = unmyelinated(
                    y, z, diameter, L, dt=dt, freq=f_dlambda, model=model
                )
        elif model in myelinated_models:
            if n_nodes > 0:
                L = get_length_from_nodes(diameter, n_nodes)
            if Nseg_per_sec:
                axon1 = myelinated(
                    y,
                    z,
                    diameter,
                    L,
                    rec="nodes",
                    dt=dt,
                    Nseg_per_sec=Nseg_per_sec,
                    model=model,
                )
            else:
                axon1 = myelinated(
                    y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
                )
        else:
            if n_nodes > 0:
                L = get_length_from_nodes(diameter, n_nodes)
            axon1 = myelinated(
                y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
            )
        # insert test spike
        axon1.insert_I_Clamp(t_position, t_start, t_duration, t_amplitude)
        # extra-cellular stimulation
        x_elec = L * position_elec
        if model in myelinated_models:  # Align electrode with node of Ranvier
            n_nodes = len(axon1.x_nodes)
            nearest_node = np.int32(np.round(position_elec * n_nodes))
            x_elec = axon1.x_nodes[nearest_node]
            if node_shift != 0:  # shift it if needed
                if node_shift > 0:
                    if nearest_node < n_nodes - 1:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec + d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )
                else:
                    if nearest_node != 0:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec - d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )
        y_elec = dist_elec
        z_elec = z_elec
        elec_1 = point_source_electrode(x_elec, y_elec, z_elec)
        stim_1 = stimulus()
        stim_1.sinus(
            b_start, b_duration, current_amp, block_freq, dt=1 / (block_freq * 20)
        )
        stim_extra = stimulation(extra_material)
        stim_extra.add_electrode(elec_1, stim_1)
        axon1.attach_extracellular_stimulation(stim_extra)
        # simulate axon activity
        results = axon1.simulate(t_sim=t_sim)
        del axon1
        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        #filter_freq(results, "V_mem", block_freq)
        #rasterize(results, "V_mem_filtered", threshold=0)
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1  
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if results.is_blocked(AP_start=t_start, freq=block_freq) == False:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not blocked")
            amplitude_min_th = current_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike blocked")
            amplitude_max_th = current_amp
            current_amp = amplitude_max_th - delta_amp / 2

        if previous_amp == amp_max:
            current_amp = amp_min
        Niter += 1
    return current_amp


def blocking_threshold_from_axon(
    axon, block_freq=10, amp_max=2000, amp_tol=1, dt=0.005, verbose=True, **kwargs
):
    """
    Find the blocking threshold with point source approximation using binary search.

    Parameters
    ----------
    axon       : axon
            axon to stimulation
    block_freq      : float
            Blocking frequency in kHz
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    verbose         : bool
            set the verbosity of the search

    **kwargs:
            See below

    Keyword Arguments:
            amp_min        : float
                    minimum amplitude for the binary search
            t_sim          : float
                    Duration of the simulation in ms
            amp_tol_abs    : float
                    Specify an absolute amplitude tolerance for the binary search
                    in uA
            t_position     : float
                    Positition on the test electrode.
                    Relative position for unmyelinated axons.
                    Node number for myelinated axons.
            t_start        : float
                    start of the test pulse, in ms
            t_duration     : float
                    duration of the test pulse, in ms
            t_amplitude    : float
                    Amplitude of the test pulse, in nA
            b_start        : float
                    start of the blocking stimulation in ms
            b_duration     : float
                    duration of the blocking stimulation in ms

    Returns
    -------
    threshold       : float
            estimated threshold in uA
    """

    rise_warning(
        "DeprecationWarning: ",
        "blocking_threshold_point_source is obsolete use blocking_threshold_from_axon instead"
    )

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        t_sim = 40

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "t_position" in kwargs:
        t_position = kwargs.get("t_position")
    else:
        t_position = 0.05

    if "t_start" in kwargs:
        t_start = kwargs.get("t_start")
    else:
        t_start = 30

    if "t_duration" in kwargs:
        t_duration = kwargs.get("t_duration")
    else:
        t_duration = 1

    if "t_amplitude" in kwargs:
        t_amplitude = kwargs.get("t_amplitude")
    else:
        t_amplitude = 5

    if "b_start" in kwargs:
        b_start = kwargs.get("b_start")
    else:
        b_start = 3

    if "b_duration" in kwargs:
        b_duration = kwargs.get("b_duration")
    else:
        b_duration = t_sim

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol

    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing block current amplitude "
                + str(current_amp)
                + " uA"
            )

        # insert test spike
        axon.insert_I_Clamp(t_position, t_start, t_duration, t_amplitude)
        # extra-cellular stimulation

        stim_1 = stimulus()
        stim_1.sinus(
            b_start, b_duration, current_amp, block_freq, dt=1 / (block_freq * 20)
        )

        axon.change_stimulus_from_electrode(0, stim_1)

        # simulate axon activity
        results = axon.simulate(t_sim=t_sim, loaded_footprints=True)

        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if results.is_blocked(AP_start=t_start, freq=block_freq) == False:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not blocked")
            amplitude_min_th = current_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike blocked")
            amplitude_max_th = current_amp
            current_amp = amplitude_max_th - delta_amp / 2

        if previous_amp == amp_max:
            current_amp = amp_min
        Niter += 1
    return current_amp




"""
NRV-:class:`.nerve` handling.
"""

import numpy as np

from ..backend._log_interface import pass_info, rise_warning
from ..backend._NRV_Class import load_any
from ..backend._NRV_Simulable import NRV_simulable
from ..utils._stimulus import stimulus
from ._fascicles import *
from .results._nerve_results import nerve_results


#################
## Nerve class ##
#################
class nerve(NRV_simulable):
    r"""
    A nerve in NRV is defined as:
        - a list of :class:`.fascicle`
        - an `analytical` :class:`~nrv.fmod.extracellular.stimulation` or a :class:`~nrv.fmod.extracellular.FEM_stimulation` context

    A nerve can be instrumented by adding couples of electrodes+stimulus

    The extracellular context of the nerve is automatically generated from the context of its
    fascicles and an optional context added directly to the nerve.

    See Also
    --------
    :doc:`Simulables users guide</usersguide/simulables>`, :class:`Fascicle-class<._fascicles.fascicle>`, :class:`Axon-class<._axons.axon>`


    .. rubric:: Customizable Attributes:

    .. list-table::
       :widths: 10 10 10 70
       :header-rows: 1

       * - Attributes
         - Type
         - Default
         - Description
       * - ``ID``
         - ``int``
         - 0
         - Identification number of the nerve.
       *
         - ``L``
         - ``float``
         - None
         - Length of the nerve.
       *
         - ``D``
         - ``float``
         - None
         - Diameter of the nerve.
       *
         - ``y_grav_center``
         - ``float``
         - 0
         - y-position of the nerve center.
       *
         - ``z_grav_center``
         - ``float``
         - 0
         - z-position of the nerve center.
       *
         - ``postproc_label``
         - ``str``
         - None
         - Label of the axon postprocessing funtion, used for the buildin postproc functions.
       *
         - ``postproc_function``
         - ``function``
         - None
         - Axon postprocessing funtion, used for the custom postproc functions.
       *
         - ``postproc_script``
         - ``str`` | ``function``
         - None
         - Either postprocessing funtion or postprocessing funtion label, automatically set depending on the type
       *
         - ``postproc_kwargs``
         - ``dict``
         - None
         - key arguments of the postporcessing function
       *
         - ``save_results``
         - ``bool``
         - False
         - If ``True``, nerve configuration and all axon simulations results are saved in ``save_path`` directory.
       *
         - ``save_path``
         - ``str``
         - ""
         - Path of the directory where simulation results should be saved.
       *
         - ``return_parameters_only``
         - ``bool``
         - False
         - If ``True`` (and ``save_results`` also ``True``), only the parameters should be returned from the simulation.
       *
         - ``loaded_footprints``
         - ``bool``
         - False
         - If ``False``, the footprints already computed are favored over new footprint computation.
       *
         - ``verbose``
         - ``bool``
         - False
         - Plot or not.

    Note
    ----
    Customizable attributes can either be set using :meth:`nerve.set_parameters` or simply by reafecting the value of the attribute.

    Tip
    ---
    Additional simulation parameters can be changed using (:meth:`.nerve.set_axons_parameters`, :meth:`.nerve.change_stimulus_from_electrode`, ...).

    Parameters
    ----------
    length              : float
        Nerve length, in um
    diameter    : float
        Nerve diameter, in um
    Outer_D            : float
        Outer saline box diameter, in mm
    ID        : int
        Nerve unique identifier
    """

    def __init__(self, length=10_000, diameter=100, Outer_D=5, ID=0, **kwargs):
        """
        Instanciates an empty nerve.
        """
        super().__init__(**kwargs)

        # to add to a fascicle/nerve common mother class
        self.verbose = True
        self.loaded_footprints = True
        self.return_parameters_only = False
        self.save_results = False
        self.save_path = ""

        self.postproc_label = "default"
        self.postproc_function = None
        self.postproc_script = None
        self.postproc_kwargs = {}

        self.type = "nerve"
        self.L = length
        self.D = None
        self.y_grav_center = 0
        self.z_grav_center = 0
        self.Outer_D = Outer_D
        self.ID = ID

        # geometric properties
        self.y_grav_center = 0
        self.z_grav_center = 0

        # Update parameters with kwargs
        self.set_parameters(**kwargs)

        # Fascicular content
        self.fascicles_IDs = []
        self.fascicles = {}

        # Axons objects default parameters
        self.unmyelinated_param = {
            "model": "Rattay_Aberham",
            "dt": 0.001,
            "Nrec": 0,
            "Nsec": 1,
            "Nseg_per_sec": length // 25,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "v_init": None,
            "T": None,
            "threshold": -40,
        }

        self.myelinated_param = {
            "model": "MRG",
            "dt": 0.001,
            "node_shift": 0,
            "Nseg_per_sec": 3,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "rec": "nodes",
            "T": None,
            "threshold": -40,
        }

        self.set_axons_parameters(**kwargs)

        # extra-cellular stimulation
        self.extra_stim = None
        self.is_extra_stim = False
        self.added_extra_stims = []
        self.footprints = {}
        self.is_footprinted = False
        self.myelinated_nseg_per_sec = 3
        self.unmyelinated_nseg = length // 25
        # intra-cellular stimulation
        self.N_intra = 0
        self.intra_stim_position = []
        self.intra_stim_t_start = []
        self.intra_stim_duration = []
        self.intra_stim_amplitude = []
        self.intra_stim_ON = []
        ## recording mechanism
        self.record = False
        self.recorder = None
        self.is_simulated = False
        # Simulation status

        self.define_circular_contour(diameter)

    ## save/load methods
    def save(
        self,
        fname="nerve.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        fascicles_context=True,
        save=True,
        _fasc_save=False,
        blacklist=[],
    ):
        """
        Save a fascicle in a json file

        Parameters
        ----------
        fname :             str
            name of the file to save the fascicle
        extracel_context:   bool
            if True, add the extracellular context to the saving
        intracel_context:   bool
            if True, add the intracellular context to the saving
        rec_context:        bool
            if True, add the recording context to the saving
        fascicles_context:  bool
            if True, add the fascicles are fully saved
        blacklist:          list[str]
            key to exclude from saving
        save:               bool
            if false only return the dictionary

        Returns
        -------
        key_dict:           dict
            Pyhton dictionary containing all the fascicle information
        """
        bl = [i for i in blacklist]

        bl += ["postproc_function", "postproc_script"]
        if self.postproc_label not in builtin_postproc_functions:
            rise_warning(
                "custom axon postprocessing cannot be save. Be carefull to set the postproc_function again when reloading fascicle"
            )

        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += ["extra_stim", "footprints", "is_footprinted"]

        if not rec_context:
            bl += ["record", "recorder"]

        if not fascicles_context:
            bl += ["fascicles"]

        to_save = save
        return super().save(
            fname=fname,
            save=to_save,
            blacklist=bl,
            intracel_context=intracel_context,
            extracel_context=extracel_context,
            rec_context=rec_context,
            _fasc_save=_fasc_save,
        )

    def load(
        self,
        data,
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Load a fascicle configuration from a json file

        Parameters
        ----------
        fname           : str
            path to the json file describing a fascicle
        extracel_context: bool
            if True, load the extracellular context as well
        intracel_context: bool
            if True, load the intracellular context as well
        rec_context: bool
            if True, load the recording context as well
        blacklist       : list[str]
            key to exclude from loading
        """
        if type(data) == str:
            key_dic = json_load(data)
        else:
            key_dic = data

        bl = [i for i in blacklist]
        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += [
                "extra_stim",
                "footprints",
                "myelinated_nseg_per_sec",
                "unmyelinated_nseg",
                "is_footprinted",
            ]

        if not rec_context:
            bl += ["record", "recorder"]

        super().load(
            data=key_dic,
            blacklist=bl,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )

    ## Nerve property method
    @property
    def n_fasc(self):
        """
        Number of fascicles in the nerves

        Returns
        -------
        int
            Number of fascicles.
        """
        return len(self.fascicles)
    
    @property
    def n_ax(self):
        """
        Number of axons in the nerves

        Returns
        -------
        int
            Number of axons.
        """
        return self.get_n_ax()

    def get_n_ax(self, id_fasc:int|None=None)->int:
        """
        Returns the number of axons in a given fascicle or in all the nerve

        Parameters
        ----------
        id_fasc : _type_, optional
            ID of the fascicle from which the number of axons is returned; if None, number of all axons in the nerve is returned, by default None

        Returns
        -------
        int
            Number of axons.
        """
        n = 0
        if id_fasc is None:
            id_fasc = self.fascicles.keys()
        elif not np.iterable(id_fasc):
            id_fasc = [id_fasc]
        for i_fasc in id_fasc:
            if i_fasc in self.fascicles:
                fasc = self.fascicles[i_fasc]
                n += fasc.n_ax
        return n

    def set_ID(self, ID):
        """
        set the ID of the nerve

        Parameters
        ----------
        ID : int
            ID number of the nerve
        """
        self.ID = ID

    def define_length(self, L):
        """
        Set the length over the x axis of the nerve

        Parameters
        ----------
        L   : float
            length of the nerve in um.
        """
        if self.extra_stim is not None or self.N_intra > 0:
            rise_warning(
                "Modifying length of a fascicle with extra or intra cellular context can lead to error"
            )
        self.L = L
        self.set_axons_parameters(unmyelinated_nseg=self.L // 25)
        for fasc in self.fascicles.values():
            fasc.define_length(L)
        if self.is_extra_stim:
            self.extra_stim.reshape_nerve(
                Nerve_D=self.D,
                Length=self.L,
                y_c=self.y_grav_center,
                z_c=self.z_grav_center,
            )

    ## generate stereotypic Nerve
    def define_circular_contour(self, D, y_c=None, z_c=None):
        """
        Define a circular countour to the nerve

        Parameters
        ----------
        D           : float
            diameter of the circular nerve contour, in um
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        """
        self.type = "Circular"
        self.D = D
        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c
        self.A = np.pi * (D / 2) ** 2
        if self.is_extra_stim:
            pass_info("FEM nerve resized!")
            self.extra_stim.reshape_nerve(
                Nerve_D=self.D,
                Length=self.L,
                y_c=self.y_grav_center,
                z_c=self.z_grav_center,
            )

    def get_circular_contour(self):
        """
        Returns the properties of the nerve contour considered as a circle (y and z center and diameter)

        Returns
        -------
        D : float
            diameter of the contour, in um. Set to 0 if not applicable
        y : float
            y position of the contour center, in um
        z : float
            z position of the contour center, in um
        """
        y = self.y_grav_center
        z = self.z_grav_center
        D = self.D
        return D, y, z

    def fit_circular_contour(self, y_c=None, z_c=None, delta=20, Delta=None):
        """
        Define a circular countour to the nerve

        Parameters
        ----------
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        delta       : float
            distance between farest axon and contour, in um
        """
        if Delta is not None:
            rise_warning(DeprecationWarning, "please use delta instead of Delta")
            delta = Delta
        N_fasc = len(self.fascicles)
        D = 2 * delta

        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c
        if N_fasc == 0:
            pass_info("No fascicles to fit - Nerve diameter set to " + str(D) + "um")
        else:
            for fasc in self.fascicles.values():
                dist_max = (
                    fasc.D / 2
                    + (
                        (self.y_grav_center - fasc.y_grav_center) ** 2
                        + (self.z_grav_center - fasc.z_grav_center) ** 2
                    )
                    ** 0.5
                )
                D = max(D, 2 * (dist_max + delta))
        self.define_circular_contour(D, y_c=None, z_c=None)

    def define_ellipsoid_contour(self, a, b, y_c=0, z_c=0, rotate=0):
        """
        Define ellipsoidal contour
        """
        pass

    ## generate Fascicle from histology
    def import_contour(self, smthing_else):
        """
        Define contour from a file
        """
        pass

    ## Fascicles handling methods
    def get_fascicles(self, ID_only=False):
        """
        Return a :class:`._fascicle.fascicle` object from it ID in the nerve

        Parameters
        ----------
        ID_only : bool, optional
            ID of the fascicle, by default False

        Returns
        -------
        _type_
            _description_
        """
        if ID_only:
            return self.fascicles_IDs
        else:
            return self.fascicles

    def add_fascicle(self, fascicle, ID=None, y=None, z=None, **kwargs):
        """
        Add a fascicle to the list of fascicles

        Parameters
        ----------
        fascicle : object
            fascicle to add to the nerve struture
        """
        fasc = load_any(fascicle, **kwargs)
        if not is_fascicle(fasc):
            rise_warning(
                "Only fascile (nrv object, dict or file) can be added with add fascicle method: nothing added"
            )
        else:
            if ID is not None:
                fasc.set_ID(ID)
            if y is None:
                y = fasc.y_grav_center
            if z is None:
                z = fasc.z_grav_center
            fasc.translate_fascicle(y - fasc.y_grav_center, z - fasc.z_grav_center)
            if self.__check_fascicle_overlap(fasc):
                rise_warning(
                    "fascicles overlap:  fasicle " + str(fasc.ID) + " cannot be added"
                )
            else:
                if fasc.ID in self.fascicles_IDs:
                    pass_info(
                        "Fascicle "
                        + str(fasc.ID)
                        + " already in the nerve: will be replace"
                    )
                    self.fascicles[fasc.ID]
                else:
                    self.fascicles_IDs += [fasc.ID]
                    self.fascicles[fasc.ID] = fasc

                self.__merge_fascicular_context(self.fascicles[fasc.ID])
                self.fascicles[fasc.ID].define_length(self.L)

    def __check_fascicle_overlap(self, fasc: fascicle):
        for fasc_i in self.fascicles.values():
            dist_yz = (fasc.y_grav_center - fasc_i.y_grav_center) ** 2 + (
                fasc.z_grav_center - fasc_i.z_grav_center
            ) ** 2
            len_min_yz = ((fasc.D + fasc_i.D) / 2) ** 2
            if dist_yz < len_min_yz:
                return True
        return False

    def __merge_fascicular_context(self, fasc: fascicle):
        if self.is_extra_stim:
            self.extra_stim.reshape_fascicle(
                fasc.D, fasc.y_grav_center, fasc.z_grav_center, fasc.ID
            )
        if fasc.extra_stim is not None:
            self.attach_extracellular_stimulation(fasc.extra_stim)

        if fasc.recorder is not None:
            self.attach_extracellular_recorder(fasc.recorder)

        fasc.clear_context(intracel_context=False)

    ## Axons handling method
    def set_axons_parameters(
        self, unmyelinated_only=False, myelinated_only=False, **kwargs
    ):
        """
        set parameters of axons in the nerve

        Parameters
        ----------
        unmyelinated_only:      bool
            if true modification only done for unmyelinated axons parameters, by default False
        myelinated_only:        bool
            if true modification only done for myelinated axons parameters, by default False
        kwargs:
            parameters to modify (see myelinated and unmyelinated)
        """
        ## Standard keys
        for key in kwargs:
            if "model" in key:
                if not myelinated_only and kwargs[key] in unmyelinated_models:
                    self.unmyelinated_param["model"] = kwargs[key]
                elif not unmyelinated_only and kwargs[key] in myelinated_models:
                    self.myelinated_param["model"] = kwargs[key]
                else:
                    rise_warning(kwargs[key], " is not an implemented axon model")
            else:
                if not myelinated_only and key in self.unmyelinated_param:
                    self.unmyelinated_param[key] = kwargs[key]
                if not unmyelinated_only and key in self.myelinated_param:
                    self.myelinated_param[key] = kwargs[key]

        ## Specific keys
        if "unmyelinated_nseg" in kwargs:
            self.unmyelinated_param["Nseg_per_sec"] = kwargs["unmyelinated_nseg"]
        if "myelinated_nseg_per_sec" in kwargs:
            self.myelinated_param["Nseg_per_sec"] = kwargs["myelinated_nseg_per_sec"]
        if "Adelta_limit" in kwargs:
            self.Adelta_limit = kwargs["self.Adelta_limit"]

        for fasc in self.fascicles.values():
            fasc.set_axons_parameters(unmyelinated_only=True, **self.unmyelinated_param)
            fasc.set_axons_parameters(myelinated_only=True, **self.myelinated_param)

    def get_axons_parameters(self, unmyelinated_only=False, myelinated_only=False):
        """
        get parameters of axons in the nerve

        Parameters
        ----------
        unmyelinated_only:      bool
            modification will only
        myelinated_only:        bool
            modification will only
        """
        if unmyelinated_only:
            return self.unmyelinated_param
        if myelinated_only:
            return self.myelinated_param
        return self.unmyelinated_param, self.myelinated_param

    ## Representation methods
    def plot(
        self,
        axes: plt.axes,
        contour_color: str = "k",
        myel_color: str = "b",
        unmyel_color: str = "r",
        elec_color: str = "gold",
        num: bool = False,
        **kwgs,
    ):
        """
        plot the nerve in the Y-Z plane (transverse section)

        Parameters
        ----------
        axes    : matplotlib.axes
            axes of the figure to display the fascicle
        contour_color   : str
            matplotlib color string applied to the contour. Black by default
        myel_color      : str
            matplotlib color string applied to the myelinated axons. Red by default
        unmyel_color    : str
            matplotlib color string applied to the myelinated axons. Blue by default
        elec_color    : str
            matplotlib color string applied to the electrodes axons. Blue by default
        num             : bool
            if True, the index of each axon is displayed on top of the circle
        """
        ## plot contour
        axes.add_patch(
            plt.Circle(
                (self.y_grav_center, self.z_grav_center),
                self.D / 2,
                color=contour_color,
                fill=False,
                linewidth=4,
            )
        )
        for i in self.fascicles:
            fasc = self.fascicles[i]
            fasc.plot(
                axes=axes,
                contour_color="grey",
                myel_color=myel_color,
                unmyel_color=unmyel_color,
                num=num,
            )
        if self.extra_stim is not None:
            self.extra_stim.plot(axes=axes, color=elec_color, nerve_d=self.D)
        axes.set_xlim((-1.1 * self.D / 2, 1.1 * self.D / 2))
        axes.set_ylim((-1.1 * self.D / 2, 1.1 * self.D / 2))

    ## Context handling methods
    def clear_context(
        self, extracel_context=True, intracel_context=True, rec_context=True
    ):
        """
        Clear all stimulation and recording mecanism
        """
        if extracel_context:
            self.L = None
            # extra-cellular stimulation
            self.extra_stim = None
            self.added_extra_stims = None
            self.footprints = {}
            self.is_footprinted = False
        if intracel_context:
            # intra-cellular stimulation
            self.N_intra = 0
            self.intra_stim_position = []
            self.intra_stim_t_start = []
            self.intra_stim_duration = []
            self.intra_stim_amplitude = []
            self.intra_stim_ON = []
        if rec_context:
            ## recording mechanism
            self.record = False
            self.recorder = None
        self.is_simulated = False

    # Intra cellular
    def insert_I_Clamp(self, position, t_start, duration, amplitude, ax_list=None):
        """
        Insert a IC clamp stimulation

        Parameters
        ----------
        position    : float
            relative position over the fascicle. Note that all thin myelinated and myelinated
            will be stimulated in the nearest node of Ranvier around the clamp specified position
        t_start     : float
            starting time, in ms
        duration    : float
            duration of the pulse, in ms
        amplitude   : float
            amplitude of the pulse (nA)
        ax_list     : list, array, np.array
            list of axons to insert the clamp on, if None, all axons are stimulated
        """
        for fasc in self.fascicles.values():
            fasc.insert_I_Clamp(
                position=position,
                t_start=t_start,
                duration=duration,
                amplitude=amplitude,
                ax_list=ax_list,
            )
        self.N_intra += 1

    def clear_I_clamp(self):
        """
        Clear any I-clamp attached to the nerve
        """
        for fasc in self.fascicles.values():
            fasc.clear_I_clamp()
        self.N_intra = 0  

    # Extracellular
    def attach_extracellular_stimulation(self, stimulation): #: FEM_stimulation):
        """
        attach a extracellular context of simulation for an axon.

        Parameters
        ----------
            stimulation  : stimulation object, see Extracellular.stimulation help for more details
        """
        if not self.is_extra_stim:
            self.extra_stim = FEM_stimulation(
                endo_mat=stimulation.endoneurium,
                peri_mat=stimulation.perineurium,
                epi_mat=stimulation.epineurium,
                ext_mat=stimulation.external_material,
            )

            self.extra_stim.reshape_outerBox(self.Outer_D)
            self.extra_stim.reshape_nerve(
                Nerve_D=self.D,
                Length=self.L,
                y_c=self.y_grav_center,
                z_c=self.z_grav_center,
            )
            self.extra_stim.remove_fascicles()
            for fasc in self.fascicles.values():
                self.extra_stim.reshape_fascicle(
                    Fascicle_D=fasc.D,
                    y_c=fasc.y_grav_center,
                    z_c=fasc.z_grav_center,
                    ID=fasc.ID,
                )
            self.is_extra_stim = True

        for i, elec in enumerate(stimulation.electrodes):
            _overlap = self.__check_perineurium_electrode_overlap(elec)
            if _overlap == -1:
                self.extra_stim.add_electrode(elec, stimulation.stimuli[i])
                for fasc in self.fascicles.values():
                    fasc.remove_axons_electrode_overlap(elec)
            else:
                rise_warning(
                    f"Electrode {i} ovelerlap with the perineurium the fascicle {_overlap}"
                )

    def change_stimulus_from_electrode(self, ID_elec, stimulus):
        """
        Change the stimulus of the ID_elec electrods

        Parameters
        ----------
            ID_elec  : int
                ID of the electrode which should be changed
            stimulus  : stimulus
                New stimulus to set
        """
        if self.extra_stim is not None:
            self.extra_stim.change_stimulus_from_electrode(ID_elec, stimulus)
        else:
            rise_warning("Cannot be changed: no extrastim in the nerve")

    def __check_perineurium_electrode_overlap(self, electrode: electrode) -> int:
        """
        Check if an electrode overlap with one fascicle perineurium of the nerve.

        Paramters
        ---------
        electrode   : electrode
            electrode to check

        Returns
        -------
        int
            ID of the fascicle overlapping, -1 if no overlap
        """
        y, z, D = 0, 0, 0
        # CUFF electrodes do not affect intrafascicular state
        if not is_CUFF_electrode(electrode):
            y = electrode.y
            z = electrode.z
            if is_LIFE_electrode(electrode):
                D = electrode.D
            for fasc in self.fascicles.values():
                dist = np.sqrt(
                    (fasc.y_grav_center - y) ** 2 + (fasc.z_grav_center - z) ** 2
                ) - (fasc.D / 2)
                if abs(dist) < D:
                    return fasc.ID
        return -1

    # RECORDING MECHANIMS
    def attach_extracellular_recorder(self, rec: recorder):
        """
        attach an extracellular recorder to the axon

        Parameters
        ----------
        rec     : recorder object
            see Recording.recorder help for more details
        """
        if is_recorder(rec):
            if self.recorder is None:
                self.recorder = rec
            else:
                for rec_p in rec.recording_points:
                    self.recorder.add_recording_point(rec_p)
            self.record = True

    def shut_recorder_down(self):
        """
        Shuts down the recorder locally
        """
        self.record = False
        for fasc in self.fascicles.values():
            fasc.shut_recorder_down()

    def __set_fascicles_context(self):
        """ """
        for fasc in self.fascicles.values():
            if self.extra_stim is not None:
                fasc.attach_extracellular_stimulation(self.extra_stim)
            if self.recorder is not None:
                fasc.attach_extracellular_recorder(self.recorder)


    def __set_fascicles_simulation_parameters(self):
        self.__set_fascicles_context()
        for fasc in self.fascicles.values():
            fasc.t_sim = self.t_sim
            fasc.record_V_mem = self.record_V_mem
            fasc.record_I_mem = self.record_I_mem
            fasc.record_I_ions = self.record_I_ions
            fasc.record_g_mem = self.record_g_mem
            fasc.record_g_ions = self.record_g_ions
            fasc.record_particles = self.record_particles

            fasc.postproc_script = self.postproc_script
            fasc.postproc_label = self.postproc_label
            fasc.postproc_function = self.postproc_function
            fasc.save_results = self.save_results
            fasc.return_parameters_only = self.return_parameters_only
            fasc.verbose = self.verbose

    # Footprint and mesh
    @property
    def __footprint_to_compute(self):
        """
        returns True only if loaded footprint are selected AND footprint exist
        """
        return not (self.is_footprinted and self.loaded_footprints)

    def compute_electrodes_footprints(self, **kwargs):
        """
        compute electrodes footprints
        """
        if not self.is_footprinted:
            self.__set_fascicles_context()
            self.set_axons_parameters(**kwargs)
            if self.is_extra_stim:
                if is_FEM_extra_stim(self.extra_stim):
                    self.extra_stim.run_model()
                for fasc in self.fascicles.values():
                    fasc.compute_electrodes_footprints()
            self.is_footprinted = True

    ## SIMULATION
    def simulate(
        self,
        **kwargs,
    ) -> nerve_results:
        """
        Simulate the nerve with the proposed extracellular context. Top level method for large
        scale neural simulation.

        Parameters
        ----------

        Warning
        -------
        calling this method can result in long processing time, even with large computational ressources.
        Keep aware of what is really implemented, ensure configuration and simulation protocol is correctly designed.
        """
        if self.verbose:
            pass_info("Starting nerve simulation")
        nerve_sim = super().simulate(**kwargs)
        self.__set_fascicles_simulation_parameters()

        if (
            self.save_path
        ):  # LR: Force folder creation if any save_path is specified --> usefull for some PP functions (ex: scatter_plot)
            folder_name = self.save_path + "Nerve_" + str(self.ID) + "/"
            create_folder(folder_name)

        if self.save_results:
            folder_name = self.save_path + "Nerve_" + str(self.ID) + "/"
            config_filename = folder_name + "/00_Nerve_config.json"
            self.save(config_filename, fascicles_context=False)

        # run FEM model
        if self.verbose:
            pass_info("...computing electrodes footprint")
        

        if self.__footprint_to_compute and self.has_FEM_extracel:
            self.compute_electrodes_footprints()
            self.loaded_footprints = True

        # Simulate all fascicles
        fasc_kwargs = kwargs
        if self.save_path:
            fasc_kwargs["save_path"] = folder_name
        if self.verbose:
            i_pbar = 1
 
        for fasc in self.fascicles.values():
            if self.verbose:
                fasc_kwargs["pbar_label"] = f"fascicle {i_pbar}/{self.n_fasc}"
                i_pbar += 1
            nerve_sim["fascicle" + str(fasc.ID)] = fasc.simulate(
                in_nerve=True,
                **fasc_kwargs,
            )
        if self.verbose:
            pass_info("...Done!")
        # dirty hack to force NRV_class type when saved
        if "extra_stim" in nerve_sim:
            nerve_sim["extra_stim"] = load_any(nerve_sim["extra_stim"])
        if self.record:  # recorder not saved in result !!BUG
            nerve_sim["recorder"] = self.recorder
        self.is_simulated = True
        return nerve_sim



""" backend: code for IO, machine and classes handling

the backend sub-package is dedicated to handling the magic behind all NRV
computations. The code developed here is not meant to be used directly by the
end-user, but internally to develop new functionality or maintain other
sub-packages.

"""

from ._config import nrv_config
from ._file_handler import (
    is_iterable,
    rmv_ext,
    generate_new_fname,
    create_folder,
    check_json_fname,
    json_dump,
    json_load,
    load_dxf_file,
    NRV_Encoder,
)
from ._inouts import (
    set_attributes,
    check_function_kwargs,
    function_to_str,
    str_to_function,
)
from ._log_interface import (
    init_reporter,
    set_log_level,
    rise_error,
    rise_warning,
    pass_info,
    pass_debug_info,
    progression_popup,
    prompt_debug,
    clear_prompt_line,
    bcolors,
)
from ._NRV_Class import (
    is_NRV_class,
    is_NRV_class_list,
    is_NRV_class_dict,
    is_NRV_dict,
    is_NRV_dict_list,
    is_NRV_dict_dict,
    is_NRV_object_dict,
    is_empty_iterable,
    load_any,
    NRV_class,
)
from ._NRV_Results import generate_results, NRV_results, sim_results
from ._NRV_Simulable import is_NRV_simulable, simulable, NRV_simulable
from ._NRV_Singleton import NRV_singleton
from ._parameters import nrv_parameters

submodules = []

classes = [
    "nrv_config",
    "NRV_Encoder",
    "bcolors",
    "NRV_class",
    "NRV_results",
    "sim_results",
    "NRV_simulable",
    "NRV_singleton",
    "nrv_parameters",
]

functions = [
    "NeuronCompile",
    "is_iterable",
    "rmv_ext",
    "generate_new_fname",
    "create_folder",
    "check_json_fname",
    "json_dump",
    "json_load",
    "load_dxf_file",
    "set_attributes",
    "check_function_kwargs",
    "function_to_str",
    "str_to_function",
    "init_reporter",
    "set_log_level",
    "rise_error",
    "rise_warning",
    "pass_info",
    "pass_debug_info",
    "progression_popup",
    "prompt_debug",
    "clear_prompt_line",
    "is_NRV_class",
    "is_NRV_class_list",
    "is_NRV_class_dict",
    "is_NRV_dict",
    "is_NRV_dict_list",
    "is_NRV_dict_dict",
    "is_NRV_object_dict",
    "is_empty_iterable",
    "load_any",
    "generate_results",
    "is_NRV_simulable",
    "simulable",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



"""
Miscellaneous functions usefull functions used in and to use with NRV
"""
import math
import numpy as np
from numpy.typing import NDArray
from pandas import DataFrame


#############################
## miscellaneous functions ##
#############################
def distance_point2point(x_1, y_1, x_2=0, y_2=0):
    """
    Computes the distance between a point (x_p,y_p) and a line defined as y=a*x+b

    Parameters
    ----------
    x_1 : float
        first point x coordinate,
    y_1 : float
        firstpoint y coordinate,
    x_2 : float
        second point x coordinate, by default 0
    y_2 : float
        second point y coordinate, , by default 0

    Returns
    --------
    d : float
        distance between the point and the orthogonal projection of (x_p,y_p) on it
    """
    d = ((x_1 - x_2) ** 2 + (y_1 - y_2) ** 2) ** 0.5
    return d


def distance_point2line(x_p, y_p, a, b):
    """
    Computes the distance between a point (x_p,y_p) and a line defined as y=a*x+b

    Parameters
    ----------
    x_p : float
        point x coordinate,
    y_p : float
        point y coordinate,
    a   : float
        line direction coeefficient
    b   : float
        line y for x = 0

    Returns
    --------
    d : float
        distance between the point and the orthogonal projection of (x_p,y_p) on it
    """
    d = np.abs(a * x_p - y_p + b) / (np.sqrt(a**2 + 1))
    return d


def nearest_idx(array: NDArray, val: float) -> int:
    """
    Return index of neareast value

    Parameters
    ----------
    array : NDArray
        array of values
    val : float
        neareast value to find index

    Returns
    -------
    int
        index of the neareast value
    """
    return (np.absolute(array - val)).argmin()


def nearest_greater_idx(array: NDArray, val: float) -> int:
    """
    Return index of neareast greater value

    Parameters
    ----------
    array : NDArray
        array of values
    val : float
        neareast greater value to find index

    Returns
    -------
    int
        index of the neareast greater value
    """
    diff = array - val
    mask = np.ma.MaskedArray(diff, diff < 0)
    return mask.argmin()


def in_tol(test: float, ref: float, tol: float = 0.1) -> bool:
    """
    Check if a value is equal to a value +- a tol (excluded).

    Parameters
    ----------
    test : float
        Value to test
    ref : float
        Reference value
    tol : float, optional
        Tolerance, expressed as a proportion of ref. By default 0.1

    Returns
    -------
    bool
        True if within tolerance, else False.
    """
    up_bound = ref * (1 + tol)
    down_bound = ref * (1 - tol)
    return np.abs(test) < up_bound and np.abs(test) > down_bound


####################################################
############# nmod related functions ###############
####################################################
def get_perineurial_thickness(fasc_d: float, nerve_type: str = "default") -> float:
    """
    Return a fascicle's perineurium thickness from its diameter.
    Relationship extract from [1]

    Parameters
    ----------
    fasc_d: float
        diameter of the fascicle
    nerve_type: str
        type of nerve containing the fascicle, possible types:

                        - "default"
                        - "sciatic"
                        - "medial popliteal"
                        - "lateral popliteal"
                        - "ulnar"
                        - "radial"
                        - "median"

    Return
    ------
    float:
        thickness of the perineurium

    Note
    ----
    [1] Y. Grinberg, M. A. Schiefer, D. J. Tyler, and K. J. Gustafson, “Fascicular Perineurium Thickness, Size, and Position Affect Model Predictions of Neural Excitation,” IEEE Trans. Neural Syst. Rehabil. Eng., vol. 16, no. 6, pp. 572–581, Dec. 2008.
    """
    thpc = 0.03
    if nerve_type.lower() == "sciatic":
        thpc = 0.033
    elif nerve_type.lower() == "medial popliteal":
        thpc = 0.026
    elif nerve_type.lower() == "lateral popliteal":
        thpc = 0.028
    elif nerve_type.lower() == "ulnar":
        thpc = 0.026
    elif nerve_type.lower() == "radial":
        thpc = 0.03
    elif nerve_type.lower() == "median":
        thpc = 0.033

    return fasc_d * thpc


MRG_data = DataFrame(data={
"fiberD": np.asarray([1, 2, 5.7, 7.3, 8.7, 10.0, 11.5, 12.8, 14.0, 15.0, 16.0]),
"g": np.asarray(
    [0.565, 0.585, 0.605, 0.630, 0.661, 0.690, 0.700, 0.719, 0.739, 0.767, 0.791]
),
"axonD": np.asarray([0.8, 1.6, 3.4, 4.6, 5.8, 6.9, 8.1, 9.2, 10.4, 11.5, 12.7]),
"nodeD": np.asarray([0.7, 1.4, 1.9, 2.4, 2.8, 3.3, 3.7, 4.2, 4.7, 5.0, 5.5]),
"paraD1": np.asarray([0.7, 1.4, 1.9, 2.4, 2.8, 3.3, 3.7, 4.2, 4.7, 5.0, 5.5]),
"paraD2": np.asarray([0.8, 1.6, 3.4, 4.6, 5.8, 6.9, 8.1, 9.2, 10.4, 11.5, 12.7]),
"deltax": np.asarray([100, 200, 500, 750, 1000, 1150, 1250, 1350, 1400, 1450, 1500]),
"paralength2": np.asarray([5, 10, 35, 38, 40, 46, 50, 54, 56, 58, 60]),
"nl": np.asarray([15, 20, 80, 100, 110, 120, 130, 135, 140, 145, 150])
})

def get_MRG_parameters(diameter:float|NDArray, fit_all:bool=False)->tuple[8]:
    """
    Compute the MRG geometrical parameters from interpolation of original data [1]

    Note
    ----
    The data used is stored in the ``MRG_data`` DataFrame, so it can be printed as follows
        >>> print(nrv.MRG_data)

    Parameters
    ----------
    diameter    : float
        diameter of the unmylinated axon to implement, in um
    fit_all     : bool
        if False, for diameters included in ``MRG_data``, original data are used without interpollation


    Returns
    -------
    g           : float
    axonD       : float
    nodeD       : float
    paraD1      : float
    paraD2      : float
    deltax      : float
    paralength2 : float
    nl          : float

    Note
    ----
    [1] McIntyre CC, Richardson AG, and Grill WM. Modeling the excitability of mammalian nerve fibers: influence of afterpotentials on the recovery cycle. Journal of Neurophysiology 87:995-1006, 2002.
    """
    MRG_fiberD = MRG_data.fiberD.to_numpy()
    MRG_g = MRG_data.g.to_numpy()
    MRG_axonD = MRG_data.axonD.to_numpy()
    MRG_nodeD = MRG_data.nodeD.to_numpy()
    MRG_paraD1 = MRG_data.paraD1.to_numpy()
    MRG_paraD2 = MRG_data.paraD2.to_numpy()
    MRG_deltax = MRG_data.deltax.to_numpy()
    MRG_paralength2 = MRG_data.paralength2.to_numpy()
    MRG_nl = MRG_data.nl.to_numpy()

    fit_all |= isinstance(diameter, np.ndarray)
    if not fit_all and diameter in MRG_fiberD:
        index = np.where(MRG_fiberD == diameter)[0]
        g = MRG_g[index]
        axonD = MRG_axonD[index]
        nodeD = MRG_nodeD[index]
        paraD1 = MRG_paraD1[index]
        paraD2 = MRG_paraD2[index]
        deltax = MRG_deltax[index]
        paralength2 = MRG_paralength2[index]
        nl = MRG_nl[index]
    else:
        # create fiting polynomyals
        g_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_g, 3))
        axonD_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_axonD, 3))
        nodeD_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_nodeD, 3))
        paraD1_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_paraD1, 3))
        paraD2_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_paraD2, 3))

        # outside of the MRG originla limit, take 1st order approx,
        paralength2_poly_OoB = np.poly1d(np.polyfit(MRG_fiberD, MRG_paralength2, 3))
        deltax_poly_OoB = np.poly1d(np.polyfit(MRG_fiberD, MRG_deltax, 2))

        # try to fit a bit better
        paralength2_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_paralength2, 5))
        deltax_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_deltax, 5))
        nl_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_nl, 3))

        # evaluate for the requested diameter
        g = g_poly(diameter)
        axonD = axonD_poly(diameter)
        nodeD = nodeD_poly(diameter)
        paraD1 = paraD1_poly(diameter)
        paraD2 = paraD2_poly(diameter)
        nl = nl_poly(diameter)

        if isinstance(diameter, np.ndarray):
            paralength2 = np.zeros(len(diameter))
            deltax = np.zeros(len(diameter))
            I_OoB = (diameter < 1.0) | (diameter > 14.0)
            paralength2[I_OoB] = paralength2_poly_OoB(diameter[I_OoB])
            deltax[I_OoB] = deltax_poly_OoB(diameter[I_OoB])
            paralength2[~I_OoB] = paralength2_poly(diameter[~I_OoB])
            deltax[~I_OoB] = deltax_poly(diameter[~I_OoB])
            return (
                g,
                axonD,
                nodeD,
                paraD1,
                paraD2,
                deltax,
                paralength2,
                nl,
            )
        else:
            if diameter < 1.0 or diameter > 14.0:
                paralength2 = paralength2_poly_OoB(diameter)
                deltax = deltax_poly_OoB(diameter)
            else:
                paralength2 = paralength2_poly(diameter)
                deltax = deltax_poly(diameter)
    return (
        float(g),
        float(axonD),
        float(nodeD),
        float(paraD1),
        float(paraD2),
        float(deltax),
        float(paralength2),
        float(nl),
    )


def get_length_from_nodes(diameter, nodes):
    """
    Function to compute the length of a myelinated axon to get the correct number of Nodes of Ranvier
    For Myelinated models only (not compatible with A delta thin myelinated models)

    Parameters
    ----------
    diameter    : float
        diameter of the axon in um
    nodes       : int
        number of nodes in the axon

    Returns
    -------
    length      : float
        lenth of the axon with the correct number of nodes in um
    """
    MRG_fiberD = MRG_data.fiberD.to_numpy()
    MRG_deltax = MRG_data.deltax.to_numpy()
    if diameter in MRG_fiberD:
        index = np.where(MRG_fiberD == diameter)[0]
        deltax = MRG_deltax[index]
    else:
        deltax_poly = np.poly1d(np.polyfit(MRG_fiberD, MRG_deltax, 4))
        deltax = deltax_poly(diameter)
    return float(math.ceil(deltax * (nodes - 1)))



def membrane_capacitance_from_model(model):
    if model in ["MRG", "Gaines_motor", "Gaines_sensory"]:
        return 2
    if "Schild" in model:
        return 1.326291192
    return 1


def compute_complex_admitance(f: float|np.ndarray, g: float|np.ndarray, fc: float|np.ndarray) -> complex:
    """
    compute the complex admitance of a first oder system (of cutoff frequency fc and conductivity g) at a given frequency f

    Parameters
    ----------
    f : float or np.array
        frequency or frequencies for which the admitance should be computed
    g : float
        conductivity of the system
    fc : float
        cutoff frequency of the system

    Returns
    -------
    complex
        complex admitance
    """
    if isinstance(g, np.ndarray) and isinstance(f, np.ndarray):
        return g * (1 + 1j * f[:, np.newaxis]  / fc)

    return g * (1 + 1j * f / fc)



"""
NRV-fenics_materials class handling.
"""

import faulthandler
import os

import numpy as np
from dolfinx.fem import Constant, Function, functionspace
from petsc4py.PETSc import ScalarType
from ufl import as_tensor

from ....backend._file_handler import json_dump, rmv_ext
from ....backend._log_interface import rise_warning
from ....backend._NRV_Class import NRV_class
from ....backend._parameters import parameters
from ....utils._nrv_function import nrv_interp
from ..._materials import is_mat, load_material
from ._f_materials import load_f_material

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()

# get the built-in material librairy
dir_path = parameters.nrv_path + "/_misc"


####################
## material class ##
####################
class fenics_material(NRV_class):
    """
    A class of materials better suited to FEM resolution with fenics.
    It allows material parameters to be updated dynamically between FEM simulations,
    which reduces calculation speed and memory consumption.

    parameters
    ----------
    mat     :material
        generate the fenics material from mat attribute
    """

    def __init__(self, mat: any = None):
        """
        initialisation of the fenics_material
        """
        super().__init__()
        self.mat = load_f_material(mat)
        self._sigma_fen = None
        self.elem = ("Discontinuous Lagrange", 1)
        self.UN = 1

    ## Save and Load mehtods
    def save(self, save=False, fname="Fenics_model.json", blacklist=[], **kwargs):
        """
        Return material as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "material.json"

        Returns
        -------
        mat_dic : dict
            dictionary containing all information
        """
        bl = [i for i in blacklist]
        bl += ["sigma_fen"]
        return super().save(save=save, fname=fname, blacklist=bl, **kwargs)

    def update_mat(self, mat):
        self.mat = load_f_material(mat)
        self.update_fenics_sigma()

    @property
    def sigma_fen(self):
        return self._sigma_fen

    @property
    def sigma(self):
        """
        conductivity of the material scaled by `UN` factor.

        Warning
        -------
        `fenics_material.sigma` and `fenics_material.mat.sigma` are not always equal
        as `fenics_material.sigma` is scaled by `UN`
        """
        return self.mat.sigma * self.UN

    def get_fenics_sigma(
        self, domain, elem=("Discontinuous Lagrange", 1), UN=1, id=None
    ):
        """
        Returns fenicsx compatible sigma
        """
        self.update_fenics_sigma(domain, elem=elem, UN=UN, id=id)
        return self._sigma_fen

    def update_fenics_sigma(self, domain=None, elem=None, UN=None, id=None):
        """
        Update sigma value of a material in fenics FEM simulation.
        Allow to reuse the same material when FEM_simulation object is solved multiple times.

        Parameters
        ----------
        domain      : dolfinx.mesh.Mesh or None
            mesh domain of the FEM simulation
        elem        : tupple (str, int)
            element type, element order, by default None
        UN          : int or None
            unit factor to use for to scale the material value,
            if None, keep the UN attribute of the instance
        id          :
            ID to use to name the ufl object
        """
        if id is None:
            name = ""
        else:
            name = "_" + str(id)
        if elem is not None:
            self.elem = elem
        if UN is not None:
            self.UN = UN
        # First declaration in fenics model (Sigma value has not been set into model)
        _sig = self.sigma
        if self._sigma_fen is None:
            # isotropic material: sigma set as a cst
            if self.mat.is_isotropic():
                self._sigma_fen = Constant(domain, ScalarType(_sig))
            # constant anisotropic material: sigma set as a tensor
            elif not self.mat.is_function_defined():
                self._sigma_fen = as_tensor(
                    [
                        [_sig[0], 0, 0],
                        [0, _sig[1], 0],
                        [0, 0, _sig[2]],
                    ]
                )
            # function defined anisotropic material: sigma set as a product of a
            # function and a constant (unit coeficient)
            else:
                Q = functionspace(domain, self.elem)
                self._sigma_fen = Function(Q)
                self._sigma_fen.interpolate(_sig)
                self._sigma_fen.name = "f" + name
        # Sigma value has already been set into fencis FEM model
        else:
            # isotropic material: sigma set as a constant
            if self.mat.is_isotropic():
                self._sigma_fen.value = _sig
            # constant anisotropic material: sigma set as a tensor
            elif not self.mat.is_function_defined():
                self._sigma_fen = as_tensor(
                    [
                        [_sig[0], 0, 0],
                        [0, _sig[1], 0],
                        [0, 0, _sig[2]],
                    ]
                )
            else:
                self._sigma_fen.interpolate(_sig)



"""
NRV-:class:`.NRV_simulable` handling.

A generic class for all NRV simulable classes (:class:`~nrv.nmod.nerve`, :class:`~nrv.nmod.fascicle`, :class:`~nrv.nmod.myelinated`)
"""

from ._NRV_Class import NRV_class
from ._NRV_Results import generate_results, sim_results


def is_NRV_simulable(x):
    """
    Check if the object x is a :class:`.NRV_simulable`.

    Parameters
    ----------
    x   : any
        object to check.

    Returns
    -------
    bool
    """
    return isinstance(x, NRV_simulable)


def simulable(x):
    """
    Check if the object x is a :class:`.NRV_simulable`.

    Parameters
    ----------
    x   : any
        object to check.

    Returns
    -------
    bool
    """
    return isinstance(x, NRV_simulable)


class NRV_simulable(NRV_class):
    """
    Generic class for all NRV simulable classes (:class:`~nrv.nmod.nerve`, :class:`~nrv.nmod.fascicle`, :class:`~nrv.nmod.myelinated`).
    This class gather commun features to all the simulable object, in particular the :func:`.NRV_simulable.simulate` method.

    Note
    ----
    -   All NRV_simulable instance are callable object. When called, the instance :func:`.NRV_simulable.simulate` method is called.

    Parameters
    ----------
    t_sim : float
        duration of the simulation in `ms`
    dt : float
        Time step of the simulation in `ms`
    record_V_mem : bool
        if true, the membrane voltage of each cell is recorded, by default True
    record_I_mem : bool
        if true, the membrane current of each cell is recorded, by default False
    record_I_ions : bool
        if true, the ionic currents of each cell are recorded, by default False
    record_particles : bool
        if true, the marticule states of each cell are recorded, by default False
    record_g_mem : bool
        if true, the membrane coductivity of each cell is recorded, by default False
    record_g_ions : bool
        if true, the ionic conductivities of each cell are recorded, by default False
    loaded_footprints : bool
        Dictionnary composed of extracellular footprint array, the keys are int value
        of the corresponding electrode ID, if None, footprints calculated during the simulation,
        set to None by default

    Note
    ----
    -   All the above parameters can either be set when the instance is initialized, or later in the script.
    """

    def __init__(
        self,
        t_sim=2e1,
        dt=0.001,
        record_V_mem=True,
        record_I_mem=False,
        record_I_ions=False,
        record_particles=False,
        record_g_mem=False,
        record_g_ions=False,
        loaded_footprints=None,
        **kwargs
    ):
        super().__init__()
        self.t_sim = t_sim
        self.record_V_mem = record_V_mem
        self.record_I_mem = record_I_mem
        self.record_I_ions = record_I_ions
        self.record_particles = record_particles
        self.record_g_mem = record_g_mem
        self.record_g_ions = record_g_ions
        self.loaded_footprints = loaded_footprints

        self.dt = dt

    @property
    def has_FEM_extracel(self) -> bool:
        #return self.extracel_status() and issubclass(self.extra_stim.nrv_type == "FEM_stimulation")
        return self.extracel_status() and self.extra_stim.nrv_type == "FEM_stimulation"

    def extracel_status(self):
        """
        Check if an extracellular context is attached to the instance

        Returns
        -------
        bool
        """
        if "extra_stim" in self.__dict__:
            if self.__dict__["extra_stim"] is not None:
                return True
        else:
            return False

    def intracel_status(self):
        """
        Check if an intracellular context is attached to the instance

        Returns
        -------
        bool
        """
        if "intra_current_stim" in self.__dict__:
            if (
                self.__dict__["intra_current_stim"] != []
                and self.__dict__["intra_current_stim"] is not None
            ):
                return True
        elif "intra_voltage_stim" in self.__dict__:
            if self.__dict__["intra_voltage_stim"] is not None:
                return True
        else:
            return False

    def rec_status(self):
        """
        Check if a recording context is attached to the instance

        Returns
        -------
        bool
        """
        if "recorder" in self.__dict__:
            if self.__dict__["recorder"] is not None:
                return True
        else:
            return False

    def __call__(self, **kwds):
        return self.simulate(**kwds)

    def simulate(self, **kwargs) -> sim_results:
        """
        Generic start of the simulate method. At this level the method does only two things:

        -   update instance attribues from the kwargs
        -   generate the sim_results dictionary

        Parameters
        ----------
        **kwargs
            Key arguments containing one or multiple parameters to set.

        Returns
        -------
        sim_results:
            Empty results containing only the simulation parameters.
        """
        self.set_parameters(**kwargs)
        context = self.save(
            save=False,
            extracel_context=self.extracel_status(),
            intracel_context=self.intracel_status(),
            rec_context=self.rec_status(),
        )
        results = generate_results(context)
        return results
